@@ -1,6 +1,13 @@
+/*	$OpenBSD: common.c,v 1.9 1998/06/28 02:22:13 angelos Exp $	*/
+
 /*
- * Copyright (c) 1983 Regents of the University of California.
- * All rights reserved.
+ * Copyright (c) 1983, 1993
+ *	The Regents of the University of California.  All rights reserved.
+ * (c) UNIX System Laboratories, Inc.
+ * All or some portions of this file are derived from material licensed
+ * to the University of California by American Telephone and Telegraph
+ * Co. or Unix System Laboratories, Inc. and are reproduced herein with
+ * the permission of UNIX System Laboratories, Inc.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -32,72 +39,105 @@
  */
 
 #ifndef lint
-/*static char sccsid[] = "from: @(#)common.c	5.7 (Berkeley) 3/2/91";*/
-static char rcsid[] = "$Id: common.c,v 1.3 1993/12/08 00:47:01 jtc Exp $";
+#if 0
+static char sccsid[] = "@(#)common.c	8.5 (Berkeley) 4/28/95";
+#else
+static char rcsid[] = "$OpenBSD: common.c,v 1.9 1998/06/28 02:22:13 angelos Exp $";
+#endif
 #endif /* not lint */
 
+#include <sys/param.h>
+#include <sys/stat.h>
+#include <sys/time.h>
+
+#include <sys/socket.h>
+#include <netinet/in.h>
+#include <arpa/inet.h>
+#include <netdb.h>
+
+#include <dirent.h>
+#include <errno.h>
+#include <unistd.h>
+#include <stdlib.h>
+#include <stdio.h>
+#include <string.h>
+#include "lp.h"
+#include "pathnames.h"
+
 /*
  * Routines and data common to all the line printer functions.
  */
 
-#include "lp.h"
-
-int	DU;		/* daeomon user-id */
-int	MX;		/* maximum number of blocks to copy */
-int	MC;		/* maximum number of copies allowed */
+char	*AF;		/* accounting file */
+long	 BR;		/* baud rate if lp is a tty */
+char	*CF;		/* name of cifplot filter (per job) */
+char	*DF;		/* name of tex filter (per job) */
+long	 DU;		/* daeomon user-id */
+long	 FC;		/* flags to clear if lp is a tty */
+char	*FF;		/* form feed string */
+long	 FS;		/* flags to set if lp is a tty */
+char	*GF;		/* name of graph(1G) filter (per job) */
+long	 HL;		/* print header last */
+char	*IF;		/* name of input filter (created per job) */
+char	*LF;		/* log file for error messages */
+char	*LO;		/* lock file name */
 char	*LP;		/* line printer device name */
+long	 MC;		/* maximum number of copies allowed */
+char	*MS;		/* stty flags to set if lp is a tty */
+long	 MX;		/* maximum number of blocks to copy */
+char	*NF;		/* name of ditroff filter (per job) */
+char	*OF;		/* name of output filter (created once) */
+char	*PF;		/* name of vrast filter (per job) */
+long	 PL;		/* page length */
+long	 PW;		/* page width */
+long	 PX;		/* page width in pixels */
+long	 PY;		/* page length in pixels */
+char	*RF;		/* name of fortran text filter (per job) */
+char    *RG;		/* resricted group */
 char	*RM;		/* remote machine name */
 char	*RP;		/* remote printer name */
-char	*LO;		/* lock file name */
-char	*ST;		/* status file name */
+long	 RS;		/* restricted to those with local accounts */
+long	 RW;		/* open LP for reading and writing */
+long	 SB;		/* short banner instead of normal header */
+long	 SC;		/* suppress multiple copies */
 char	*SD;		/* spool directory */
-char	*AF;		/* accounting file */
-char	*LF;		/* log file for error messages */
-char	*OF;		/* name of output filter (created once) */
-char	*IF;		/* name of input filter (created per job) */
-char	*RF;		/* name of fortran text filter (per job) */
+long	 SF;		/* suppress FF on each print job */
+long	 SH;		/* suppress header page */
+char	*ST;		/* status file name */
 char	*TF;		/* name of troff filter (per job) */
-char	*NF;		/* name of ditroff filter (per job) */
-char	*DF;		/* name of tex filter (per job) */
-char	*GF;		/* name of graph(1G) filter (per job) */
-char	*VF;		/* name of vplot filter (per job) */
-char	*CF;		/* name of cifplot filter (per job) */
-char	*PF;		/* name of vrast filter (per job) */
-char	*FF;		/* form feed string */
 char	*TR;		/* trailer string to be output when Q empties */
-short	SH;		/* suppress header page */
-short	SB;		/* short banner instead of normal header */
-short	HL;		/* print header last */
-short	RW;		/* open LP for reading and writing */
-short	PW;		/* page width */
-short	PL;		/* page length */
-short	PX;		/* page width in pixels */
-short	PY;		/* page length in pixels */
-short	BR;		/* baud rate if lp is a tty */
-int	FC;		/* flags to clear if lp is a tty */
-int	FS;		/* flags to set if lp is a tty */
-short	RS;		/* restricted to those with local accounts */
+char	*VF;		/* name of vplot filter (per job) */
+long	 XC;		/* flags to clear for local mode */
+long	 XS;		/* flags to set for local mode */
 
 char	line[BUFSIZ];
-char	pbuf[BUFSIZ/2];	/* buffer for printcap strings */
-char	*bp = pbuf;	/* pointer into pbuf for pgetent() */
+char	*bp;		/* pointer into printcap buffer. */
 char	*name;		/* program name */
 char	*printer;	/* printer name */
-char	host[32];	/* host machine name */
+			/* host machine name */
+char	host[MAXHOSTNAMELEN];
 char	*from = host;	/* client's machine name */
-int	sendtorem;	/* are we sending to a remote? */
+int	remote;		/* true if sending files to a remote host */
+char	*printcapdb[2] = { _PATH_PRINTCAP, 0 };
+
+extern uid_t	uid, euid;
+
+static int compar __P((const void *, const void *));
 
 /*
- * Create a connection to the remote printer server.
+ * Create a TCP connection to host "rhost" at port "rport".
+ * If rport == 0, then use the printer service port.
  * Most of this code comes from rcmd.c.
  */
-getport(rhost)
+int
+getport(rhost, rport)
 	char *rhost;
+	int rport;
 {
 	struct hostent *hp;
 	struct servent *sp;
 	struct sockaddr_in sin;
-	int s, timo = 1, lport = IPPORT_RESERVED - 1;
+	int s, timo = 1, on = 1, lport = IPPORT_RESERVED - 1;
 	int err;
 
 	/*
@@ -105,22 +145,31 @@ getport(rhost)
 	 */
 	if (rhost == NULL)
 		fatal("no remote host to connect to");
-	hp = gethostbyname(rhost);
-	if (hp == NULL)
-		fatal("unknown host %s", rhost);
-	sp = getservbyname("printer", "tcp");
-	if (sp == NULL)
-		fatal("printer/tcp: unknown service");
 	bzero((char *)&sin, sizeof(sin));
-	bcopy(hp->h_addr, (caddr_t)&sin.sin_addr, hp->h_length);
-	sin.sin_family = hp->h_addrtype;
-	sin.sin_port = sp->s_port;
+	if (inet_aton(rhost, &sin.sin_addr) == 1)
+		sin.sin_family = AF_INET;
+	else {
+		hp = gethostbyname(rhost);
+		if (hp == NULL)
+			fatal("unknown host %s", rhost);
+		bcopy(hp->h_addr, (caddr_t)&sin.sin_addr, hp->h_length);
+		sin.sin_family = hp->h_addrtype;
+	}
+	if (rport == 0) {
+		sp = getservbyname("printer", "tcp");
+		if (sp == NULL)
+			fatal("printer/tcp: unknown service");
+		sin.sin_port = sp->s_port;
+	} else
+		sin.sin_port = htons(rport);
 
 	/*
 	 * Try connecting to the server.
 	 */
 retry:
+	seteuid(euid);
 	s = rresvport(&lport);
+	seteuid(uid);
 	if (s < 0)
 		return(-1);
 	if (connect(s, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
@@ -138,6 +187,9 @@ retry:
 		}
 		return(-1);
 	}
+	
+	/* Don't bother if we get an error here.  */
+	setsockopt(s, SOL_SOCKET, SO_KEEPALIVE, (char *)&on, sizeof on);
 	return(s);
 }
 
@@ -146,6 +198,7 @@ retry:
  *  new-line to null and leaves it in line.
  * Returns 0 at EOF or the number of characters read.
  */
+int
 getline(cfp)
 	FILE *cfp;
 {
@@ -153,14 +206,14 @@ getline(cfp)
 	register char *lp = line;
 	register c;
 
-	while ((c = getc(cfp)) != '\n') {
+	while ((c = getc(cfp)) != '\n' && linel+1<sizeof(line)) {
 		if (c == EOF)
 			return(0);
 		if (c == '\t') {
 			do {
 				*lp++ = ' ';
 				linel++;
-			} while ((linel & 07) != 0);
+			} while ((linel & 07) != 0 && linel+1<sizeof(line));
 			continue;
 		}
 		*lp++ = c;
@@ -175,21 +228,23 @@ getline(cfp)
  * creation time.
  * Return the number of entries and a pointer to the list.
  */
+int
 getq(namelist)
 	struct queue *(*namelist[]);
 {
-	register struct direct *d;
+	register struct dirent *d;
 	register struct queue *q, **queue;
 	register int nitems;
 	struct stat stbuf;
 	DIR *dirp;
 	int arraysz;
-	static int compar();
 
+	seteuid(euid);
 	if ((dirp = opendir(SD)) == NULL)
 		return(-1);
-	if (fstat(dirfd(dirp), &stbuf) < 0)
+	if (fstat(dirp->dd_fd, &stbuf) < 0)
 		goto errdone;
+	seteuid(uid);
 
 	/*
 	 * Estimate the array size by taking the size of the directory file
@@ -204,8 +259,10 @@ getq(namelist)
 	while ((d = readdir(dirp)) != NULL) {
 		if (d->d_name[0] != 'c' || d->d_name[1] != 'f')
 			continue;	/* daemon control files only */
+		seteuid(euid);
 		if (stat(d->d_name, &stbuf) < 0)
 			continue;	/* Doesn't exist */
+		seteuid(uid);
 		q = (struct queue *)malloc(sizeof(time_t)+strlen(d->d_name)+1);
 		if (q == NULL)
 			goto errdone;
@@ -216,8 +273,9 @@ getq(namelist)
 		 * realloc the maximum size.
 		 */
 		if (++nitems > arraysz) {
+			arraysz *= 2;
 			queue = (struct queue **)realloc((char *)queue,
-				(stbuf.st_size/12) * sizeof(struct queue *));
+				arraysz * sizeof(struct queue *));
 			if (queue == NULL)
 				goto errdone;
 		}
@@ -237,13 +295,13 @@ errdone:
 /*
  * Compare modification times.
  */
-static
+static int
 compar(p1, p2)
-	register struct queue **p1, **p2;
+	const void *p1, *p2;
 {
-	if ((*p1)->q_time < (*p2)->q_time)
+	if ((*(struct queue **)p1)->q_time < (*(struct queue **)p2)->q_time)
 		return(-1);
-	if ((*p1)->q_time > (*p2)->q_time)
+	if ((*(struct queue **)p1)->q_time > (*(struct queue **)p2)->q_time)
 		return(1);
 	return(0);
 }
@@ -258,15 +316,16 @@ checkremote()
 	char name[MAXHOSTNAMELEN];
 	register struct hostent *hp;
 	static char errbuf[128];
+	char *rp, *rp_b;
 
-	sendtorem = 0;	/* assume printer is local */
-	if (RM != (char *)NULL) {
+	remote = 0;	/* assume printer is local */
+	if (RM != NULL) {
 		/* get the official name of the local host */
 		gethostname(name, sizeof(name));
 		name[sizeof(name)-1] = '\0';
 		hp = gethostbyname(name);
 		if (hp == (struct hostent *) NULL) {
-		    (void) sprintf(errbuf,
+		    (void) snprintf(errbuf, sizeof(errbuf),
 			"unable to get official name for local machine %s",
 			name);
 		    return errbuf;
@@ -275,7 +334,7 @@ checkremote()
 		/* get the official name of RM */
 		hp = gethostbyname(RM);
 		if (hp == (struct hostent *) NULL) {
-		    (void) sprintf(errbuf,
+		    (void) snprintf(errbuf, sizeof(errbuf),
 			"unable to get official name for remote machine %s",
 			RM);
 		    return errbuf;
@@ -285,22 +344,67 @@ checkremote()
 		 * if the two hosts are not the same,
 		 * then the printer must be remote.
 		 */
-		if (strcmp(name, hp->h_name) != 0)
-			sendtorem = 1;
+		if (strcasecmp(name, hp->h_name) != 0)
+			remote = 1;
+		else if (cgetstr(bp, "rp", &rp) > 0) {
+			if (cgetent(&rp_b, printcapdb, rp) == 0) {
+				if (cgetmatch(rp_b, printer) != 0)
+					remote = 1;
+				free(rp_b);
+			} else {
+				(void) snprintf(errbuf, sizeof(errbuf),
+					"can't find (local) remote printer %s",
+					rp);
+					free(rp);
+					return errbuf;
+			}
+			free(rp);
+		}
 	}
-	return (char *)0;
+	return NULL;
 }
 
-/*VARARGS1*/
-fatal(msg, a1, a2, a3)
-	char *msg;
+/* sleep n milliseconds */
+void
+delay(n)
 {
+	struct timeval tdelay;
+
+	if (n <= 0 || n > 10000)
+		fatal("unreasonable delay period (%d)", n);
+	tdelay.tv_sec = n / 1000;
+	tdelay.tv_usec = n * 1000 % 1000000;
+	(void) select(0, (fd_set *)0, (fd_set *)0, (fd_set *)0, &tdelay);
+}
+
+#ifdef __STDC__
+#include <stdarg.h>
+#else
+#include <varargs.h>
+#endif
+
+void
+#ifdef __STDC__
+fatal(const char *msg, ...)
+#else
+fatal(msg, va_alist)
+	char *msg;
+        va_dcl
+#endif
+{
+	va_list ap;
+#ifdef __STDC__
+	va_start(ap, msg);
+#else
+	va_start(ap);
+#endif
 	if (from != host)
-		printf("%s: ", host);
-	printf("%s: ", name);
+		(void)printf("%s: ", host);
+	(void)printf("%s: ", name);
 	if (printer)
-		printf("%s: ", printer);
-	printf(msg, a1, a2, a3);
-	putchar('\n');
+		(void)printf("%s: ", printer);
+	(void)vprintf(msg, ap);
+	va_end(ap);
+	(void)putchar('\n');
 	exit(1);
 }
