@@ -1,6 +1,8 @@
+/*	$OpenBSD: lp.local.h,v 1.3 1997/06/23 23:13:56 kstailey Exp $	*/
+
 /*
- * Copyright (c) 1983 Regents of the University of California.
- * All rights reserved.
+ * Copyright (c) 1983, 1993
+ *	The Regents of the University of California.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -30,37 +32,20 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
- *	from: @(#)lp.local.h	5.5 (Berkeley) 6/1/90
- *	$Id: lp.local.h,v 1.4 1993/08/01 17:59:08 mycroft Exp $
+ *	@(#)lp.local.h	8.1 (Berkeley) 6/6/93
  */
 
 /*
  * Possibly, local parameters to the spooling system
  */
 
-/*
- * Magic number mapping for binary files, used by lpr to avoid
- *   printing objects files.
- */
-
-#include <a.out.h>
-#include <ar.h>
-
-#ifndef A_MAGIC1	/* must be a VM/UNIX system */
-#	define A_MAGIC1 OMAGIC
-#	define A_MAGIC2 NMAGIC
-#	define A_MAGIC3 ZMAGIC
-#	undef ARMAG
-#	define ARMAG	0177545
-#endif
-
 /*
  * Defaults for line printer capabilities data base
  */
 #define	DEFLP		"lp"
 #define DEFLOCK		"lock"
 #define DEFSTAT		"status"
-#define	DEFMX		1000
+#define	DEFMX		0
 #define DEFMAXCOPIES	0
 #define DEFFF		"\f"
 #define DEFWIDTH	132
