@@ -1,6 +1,8 @@
+/*	$OpenBSD: rmjob.c,v 1.9 1998/04/22 14:44:08 deraadt Exp $	*/
+
 /*
- * Copyright (c) 1983 Regents of the University of California.
- * All rights reserved.
+ * Copyright (c) 1983, 1993
+ *	The Regents of the University of California.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -32,17 +34,31 @@
  */
 
 #ifndef lint
-/*static char sccsid[] = "from: @(#)rmjob.c	5.7 (Berkeley) 6/1/90";*/
-static char rcsid[] = "$Id: rmjob.c,v 1.1 1994/05/23 09:04:58 rzsfl Exp rzsfl $";
+#if 0
+static char sccsid[] = "@(#)rmjob.c	8.2 (Berkeley) 4/28/95";
+#else
+static char rcsid[] = "$OpenBSD: rmjob.c,v 1.9 1998/04/22 14:44:08 deraadt Exp $";
+#endif
 #endif /* not lint */
 
+#include <sys/param.h>
+
+#include <signal.h>
+#include <errno.h>
+#include <dirent.h>
+#include <unistd.h>
+#include <stdlib.h>
+#include <stdio.h>
+#include <string.h>
+#include <ctype.h>
+#include "lp.h"
+#include "lp.local.h"
+#include "pathnames.h"
+
 /*
  * rmjob - remove the specified jobs from the queue.
  */
 
-#include "lp.h"
-#include "pathnames.h"
-
 /*
  * Stuff for handling lprm specifications
  */
@@ -52,34 +68,39 @@ extern int	requ[];			/* job number of spool entries */
 extern int	requests;		/* # of spool requests */
 extern char	*person;		/* name of person doing lprm */
 
-char	root[] = "root";
-int	all = 0;		/* eliminate all files (root only) */
-int	cur_daemon;		/* daemon's pid */
-char	current[40];		/* active control file name */
+static char	root[] = "root";
+static int	all = 0;		/* eliminate all files (root only) */
+static int	cur_daemon;		/* daemon's pid */
+static char	current[40];		/* active control file name */
 
-int	iscf();
+extern uid_t	uid, euid;		/* real and effective user id's */
 
+static	void	do_unlink __P((char *));
+
+void
 rmjob()
 {
 	register int i, nitems;
 	int assasinated = 0;
-	struct direct **files;
+	struct dirent **files;
 	char *cp;
 
-	if ((i = pgetent(line, printer)) < 0)
-		fatal("cannot open printer description file");
-	else if (i == 0)
+	if ((i = cgetent(&bp, printcapdb, printer)) == -2)
+		fatal("can't open printer description file");
+	else if (i == -1)
 		fatal("unknown printer");
-	if ((SD = pgetstr("sd", &bp)) == NULL)
-		SD = _PATH_DEFSPOOL;
-	if ((LO = pgetstr("lo", &bp)) == NULL)
-		LO = DEFLOCK;
-	if ((LP = pgetstr("lp", &bp)) == NULL)
+	else if (i == -3)
+		fatal("potential reference loop detected in printcap file");
+	if (cgetstr(bp, "lp", &LP) < 0)
 		LP = _PATH_DEFDEVLP;
-	if ((RP = pgetstr("rp", &bp)) == NULL)
+	if (cgetstr(bp, "rp", &RP) < 0)
 		RP = DEFLP;
-	RM = pgetstr("rm", &bp);
-	if (cp = checkremote())
+	if (cgetstr(bp, "sd", &SD) < 0)
+		SD = _PATH_DEFSPOOL;
+	if (cgetstr(bp,"lo", &LO) < 0)
+		LO = DEFLOCK;
+	cgetstr(bp, "rm", &RM);
+	if ((cp = checkremote()))
 		printf("Warning: %s\n", cp);
 
 	/*
@@ -101,10 +122,12 @@ rmjob()
 		person = root;
 	}
 
+	seteuid(euid);
 	if (chdir(SD) < 0)
 		fatal("cannot chdir to spool directory");
 	if ((nitems = scandir(".", &files, iscf, NULL)) < 0)
 		fatal("cannot access spool directory");
+	seteuid(uid);
 
 	if (nitems) {
 		/*
@@ -113,7 +136,9 @@ rmjob()
 		 *  (after which we have to restart the daemon).
 		 */
 		if (lockchk(LO) && chk(current)) {
+			seteuid(euid);
 			assasinated = kill(cur_daemon, SIGINT) == 0;
+			seteuid(uid);
 			if (!assasinated)
 				fatal("cannot kill printer daemon");
 		}
@@ -137,23 +162,27 @@ rmjob()
  *  daemon and the file name of the active spool entry.
  * Return boolean indicating existence of a lock file.
  */
+int
 lockchk(s)
 	char *s;
 {
 	register FILE *fp;
 	register int i, n;
 
-	if ((fp = fopen(s, "r")) == NULL)
+	seteuid(euid);
+	if ((fp = fopen(s, "r")) == NULL) {
 		if (errno == EACCES)
 			fatal("can't access lock file");
 		else
 			return(0);
+	}
+	seteuid(uid);
 	if (!getline(fp)) {
 		(void) fclose(fp);
 		return(0);		/* no daemon present */
 	}
 	cur_daemon = atoi(line);
-	if (kill(cur_daemon, 0) < 0) {
+	if (kill(cur_daemon, 0) < 0 && errno != EPERM) {
 		(void) fclose(fp);
 		return(0);		/* no daemon present */
 	}
@@ -172,6 +201,7 @@ lockchk(s)
 /*
  * Process a control file.
  */
+void
 process(file)
 	char *file;
 {
@@ -179,31 +209,45 @@ process(file)
 
 	if (!chk(file))
 		return;
+	seteuid(euid);
 	if ((cfp = fopen(file, "r")) == NULL)
 		fatal("cannot open %s", file);
+	seteuid(uid);
 	while (getline(cfp)) {
 		switch (line[0]) {
 		case 'U':  /* unlink associated files */
-			if (from != host)
-				printf("%s: ", host);
-			printf(unlink(line+1) ? "cannot dequeue %s\n" :
-				"%s dequeued\n", line+1);
+			if (strchr(line+1, '/') || strncmp(line+1, "df", 2))
+				break;
+			do_unlink(line+1);
 		}
 	}
 	(void) fclose(cfp);
+	do_unlink(file);
+}
+
+static void
+do_unlink(file)
+	char *file;
+{
+	int	ret;
+
 	if (from != host)
 		printf("%s: ", host);
-	printf(unlink(file) ? "cannot dequeue %s\n" : "%s dequeued\n", file);
+	seteuid(euid);
+	ret = unlink(file);
+	seteuid(uid);
+	printf(ret ? "cannot dequeue %s\n" : "%s dequeued\n", file);
 }
 
 /*
  * Do the dirty work in checking
  */
+int
 chk(file)
 	char *file;
 {
 	register int *r, n;
-	char **u, *cp, p[64], h[64];
+	register char **u, *cp;
 	FILE *cfp;
 
 	/*
@@ -212,38 +256,39 @@ chk(file)
 	if (strlen(file) < 7 || file[0] != 'c' || file[1] != 'f')
 		return(0);
 
+	if (all && (from == host || !strcmp(from, file+6)))
+		return(1);
+
 	/*
 	 * get the owner's name from the control file.
 	 */
-	p[0] = h[0] = '\0';
+	seteuid(euid);
 	if ((cfp = fopen(file, "r")) == NULL)
 		return(0);
+	seteuid(uid);
 	while (getline(cfp)) {
 		if (line[0] == 'P')
-			strcpy(p, line+1);
-		else if (line[0] == 'H')
-			strcpy(h, line+1);
+			break;
 	}
 	(void) fclose(cfp);
-
-	if (all && (from == host || !strcmp(from, h)))
-		return(1);
+	if (line[0] != 'P')
+		return(0);
 
 	if (users == 0 && requests == 0)
-		return(!strcmp(file, current) && isowner(p, file, h));
+		return(!strcmp(file, current) && isowner(line+1, file));
 	/*
 	 * Check the request list
 	 */
 	for (n = 0, cp = file+3; isdigit(*cp); )
 		n = n * 10 + (*cp++ - '0');
 	for (r = requ; r < &requ[requests]; r++)
-		if (*r == n && isowner(p, file, h))
+		if (*r == n && isowner(line+1, file))
 			return(1);
 	/*
 	 * Check to see if it's in the user list
 	 */
 	for (u = user; u < &user[users]; u++)
-		if (!strcmp(*u, p) && isowner(p, file, h))
+		if (!strcmp(*u, line+1) && isowner(line+1, file))
 			return(1);
 	return(0);
 }
@@ -254,12 +299,13 @@ chk(file)
  * files sent from the remote machine to be removed.
  * Normal users can only remove the file from where it was sent.
  */
-isowner(owner, file, h)
-	char *owner, *file, *h;
+int
+isowner(owner, file)
+	char *owner, *file;
 {
-	if (!strcmp(person, root) && (from == host || !strcmp(from, h)))
+	if (!strcmp(person, root) && (from == host || !strcmp(from, file+6)))
 		return(1);
-	if (!strcmp(person, owner) && !strcmp(from, h))
+	if (!strcmp(person, owner) && !strcmp(from, file+6))
 		return(1);
 	if (from != host)
 		printf("%s: ", host);
@@ -271,13 +317,14 @@ isowner(owner, file, h)
  * Check to see if we are sending files to a remote machine. If we are,
  * then try removing files on the remote machine.
  */
+void
 rmremote()
 {
 	register char *cp;
 	register int i, rem;
 	char buf[BUFSIZ];
 
-	if (!sendtorem)
+	if (!remote)
 		return;	/* not sending to a remote machine */
 
 	/*
@@ -286,19 +333,19 @@ rmremote()
 	 */
 	fflush(stdout);
 
-	sprintf(buf, "\5%s %s", RP, all ? "-all" : person);
-	cp = buf;
-	for (i = 0; i < users; i++) {
+	(void)snprintf(buf, sizeof(buf)-2, "\5%s %s", RP, all ? "-all" : person);
+	cp = buf + strlen(buf);
+	for (i = 0; i < users && cp-buf+1+strlen(user[i]) < sizeof buf - 2; i++) {
 		cp += strlen(cp);
 		*cp++ = ' ';
 		strcpy(cp, user[i]);
 	}
-	for (i = 0; i < requests; i++) {
+	for (i = 0; i < requests && cp-buf+10 < sizeof(buf) - 2; i++) {
 		cp += strlen(cp);
 		(void) sprintf(cp, " %d", requ[i]);
 	}
 	strcat(cp, "\n");
-	rem = getport(RM);
+	rem = getport(RM, 0);
 	if (rem < 0) {
 		if (from != host)
 			printf("%s: ", host);
@@ -316,6 +363,7 @@ rmremote()
 /*
  * Return 1 if the filename begins with 'cf'
  */
+int
 iscf(d)
 	struct dirent *d;
 {
