@@ -1,6 +1,8 @@
+/*	$OpenBSD: startdaemon.c,v 1.3 1997/01/17 16:11:38 millert Exp $	*/
+
 /*
- * Copyright (c) 1983 Regents of the University of California.
- * All rights reserved.
+ * Copyright (c) 1983, 1993, 1994
+ *	The Regents of the University of California.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -32,43 +34,65 @@
  */
 
 #ifndef lint
-/*static char sccsid[] = "from: @(#)startdaemon.c	5.7 (Berkeley) 3/2/91";*/
-static char rcsid[] = "$Id: startdaemon.c,v 1.3 1993/11/10 04:36:04 cgd Exp $";
+#if 0
+static char sccsid[] = "@(#)startdaemon.c	8.2 (Berkeley) 4/17/94";
+#else
+static char rcsid[] = "$OpenBSD: startdaemon.c,v 1.3 1997/01/17 16:11:38 millert Exp $";
+#endif
 #endif /* not lint */
 
+
+#include <sys/param.h>
+#include <sys/socket.h>
+#include <sys/un.h>
+
+#include <dirent.h>
+#include <errno.h>
+#include <stdio.h>
+#include <unistd.h>
+#include <string.h>
+#include "lp.h"
+#include "pathnames.h"
+
+extern uid_t	uid, euid;
+
+static void perr __P((char *));
+
 /*
  * Tell the printer daemon that there are new files in the spool directory.
  */
 
-#include <sys/types.h>
-#include <sys/socket.h>
-#include <sys/un.h>
-#include <stdio.h>
-#include "lp.local.h"
-#include "pathnames.h"
-
+int
 startdaemon(printer)
 	char *printer;
 {
-	struct sockaddr_un sun;
+	struct sockaddr_un un;
 	register int s, n;
 	char buf[BUFSIZ];
-	static void perr();
 
 	s = socket(AF_UNIX, SOCK_STREAM, 0);
 	if (s < 0) {
 		perr("socket");
 		return(0);
 	}
-	bzero(&sun, sizeof(sun));
-	sun.sun_family = AF_UNIX;
-	strcpy(sun.sun_path, _PATH_SOCKETNAME);
-	if (connect(s, (struct sockaddr *)&sun, strlen(sun.sun_path) + 2) < 0) {
+	memset(&un, 0, sizeof(un));
+	un.sun_family = AF_UNIX;
+	strcpy(un.sun_path, _PATH_SOCKETNAME);
+#ifndef SUN_LEN
+#define SUN_LEN(unp) (strlen((unp)->sun_path) + 2)
+#endif
+	seteuid(euid);
+	if (connect(s, (struct sockaddr *)&un, SUN_LEN(&un)) < 0) {
+		seteuid(uid);
 		perr("connect");
 		(void) close(s);
 		return(0);
 	}
-	(void) sprintf(buf, "\1%s\n", printer);
+	seteuid(uid);
+	if (snprintf(buf, sizeof buf, "\1%s\n", printer) > sizeof buf-1) {
+		close(s);
+		return (0);
+	}
 	n = strlen(buf);
 	if (write(s, buf, n) != n) {
 		perr("write");
@@ -92,9 +116,7 @@ static void
 perr(msg)
 	char *msg;
 {
-	extern int errno;
 	extern char *name;
-	char *strerror();
 
 	(void)printf("%s: %s: %s\n", name, msg, strerror(errno));
 }
