@@ -1,6 +1,9 @@
+/*	$OpenBSD: cmds.c,v 1.9 1997/07/25 18:57:24 grr Exp $	*/
+
 /*
- * Copyright (c) 1983 Regents of the University of California.
- * All rights reserved.
+ * Copyright (c) 1983, 1993
+ *	The Regents of the University of California.  All rights reserved.
+ *
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -32,22 +35,66 @@
  */
 
 #ifndef lint
-/*static char sccsid[] = "from: @(#)cmds.c	5.7 (Berkeley) 6/1/90";*/
-static char rcsid[] = "$Id: cmds.c,v 1.3 1993/12/08 00:50:06 jtc Exp $";
+static char copyright[] =
+"@(#) Copyright (c) 1983, 1993\n\
+	The Regents of the University of California.  All rights reserved.\n";
+#endif /* not lint */
+
+#ifndef lint
+#if 0
+static char sccsid[] = "@(#)cmds.c	8.2 (Berkeley) 4/28/95";
+#else
+static char rcsid[] = "$OpenBSD: cmds.c,v 1.9 1997/07/25 18:57:24 grr Exp $";
+#endif
 #endif /* not lint */
 
 /*
  * lpc -- line printer control program -- commands:
  */
 
-#include "lp.h"
+#include <sys/param.h>
 #include <sys/time.h>
+#include <sys/stat.h>
+#include <sys/file.h>
+
+#include <signal.h>
+#include <fcntl.h>
+#include <errno.h>
+#include <dirent.h>
+#include <unistd.h>
+#include <stdlib.h>
+#include <stdio.h>
+#include <ctype.h>
+#include <string.h>
+#include "lp.h"
+#include "lp.local.h"
+#include "lpc.h"
+#include "extern.h"
 #include "pathnames.h"
 
+extern uid_t	uid, euid;
+
+static void	abortpr __P((int));
+static void	cleanpr __P((void));
+static void	disablepr __P((void));
+static int	doarg __P((char *));
+static int	doselect __P((struct dirent *));
+static void	enablepr __P((void));
+static void	prstat __P((void));
+static void	putmsg __P((int, char **));
+static int	sortq __P((const void *, const void *));
+static void	startpr __P((int));
+static void	stoppr __P((void));
+static int	touch __P((struct queue *));
+static void	unlinkf __P((char *));
+static void	upstat __P((char *));
+
 /*
  * kill an existing daemon and disable printing.
  */
-xabort(argc, argv)
+void
+doabort(argc, argv)
+	int argc;
 	char *argv[];
 {
 	register int c, status;
@@ -60,10 +107,11 @@ xabort(argc, argv)
 	}
 	if (argc == 2 && !strcmp(argv[1], "all")) {
 		printer = prbuf;
-		while (getprent(line) > 0) {
+		while (cgetnext(&bp, printcapdb) > 0) {
 			cp1 = prbuf;
-			cp2 = line;
-			while ((c = *cp2++) && c != '|' && c != ':')
+			cp2 = bp;
+			while ((c = *cp2++) && c != '|' && c != ':' &&
+			    (cp1 - prbuf) < sizeof(prbuf))
 				*cp1++ = c;
 			*cp1 = '\0';
 			abortpr(1);
@@ -72,35 +120,38 @@ xabort(argc, argv)
 	}
 	while (--argc) {
 		printer = *++argv;
-		if ((status = pgetent(line, printer)) < 0) {
+		if ((status = cgetent(&bp, printcapdb, printer)) == -2) {
 			printf("cannot open printer description file\n");
 			continue;
-		} else if (status == 0) {
+		} else if (status == -1) {
 			printf("unknown printer %s\n", printer);
 			continue;
-		}
+		} else if (status == -3)
+			fatal("potential reference loop detected in printcap file");
 		abortpr(1);
 	}
 }
 
+static void
 abortpr(dis)
+	int dis;
 {
 	register FILE *fp;
 	struct stat stbuf;
 	int pid, fd;
 
-	bp = pbuf;
-	if ((SD = pgetstr("sd", &bp)) == NULL)
+	if (cgetstr(bp, "sd", &SD) == -1)
 		SD = _PATH_DEFSPOOL;
-	if ((LO = pgetstr("lo", &bp)) == NULL)
+	if (cgetstr(bp, "lo", &LO) == -1)
 		LO = DEFLOCK;
-	(void) sprintf(line, "%s/%s", SD, LO);
+	(void) snprintf(line, sizeof(line), "%s/%s", SD, LO);
 	printf("%s:\n", printer);
 
 	/*
 	 * Turn on the owner execute bit of the lock file to disable printing.
 	 */
 	if (dis) {
+		seteuid(euid);
 		if (stat(line, &stbuf) >= 0) {
 			if (chmod(line, (stbuf.st_mode & 0777) | 0100) < 0)
 				printf("\tcannot disable printing\n");
@@ -117,10 +168,10 @@ abortpr(dis)
 				printf("\tprinting disabled\n");
 				printf("\tno daemon to abort\n");
 			}
-			return;
+			goto out;
 		} else {
 			printf("\tcannot stat lock file\n");
-			return;
+			goto out;
 		}
 	}
 	/*
@@ -128,33 +179,38 @@ abortpr(dis)
 	 */
 	if ((fp = fopen(line, "r")) == NULL) {
 		printf("\tcannot open lock file\n");
-		return;
+		goto out;
 	}
 	if (!getline(fp) || flock(fileno(fp), LOCK_SH|LOCK_NB) == 0) {
 		(void) fclose(fp);	/* unlocks as well */
 		printf("\tno daemon to abort\n");
-		return;
+		goto out;
 	}
 	(void) fclose(fp);
-	if (kill(pid = atoi(line), SIGTERM) < 0)
-		printf("\tWarning: daemon (pid %d) not killed\n", pid);
-	else
+	if (kill(pid = atoi(line), SIGTERM) < 0) {
+		if (errno == ESRCH)
+			printf("\tno daemon to abort\n");
+		else
+			printf("\tWarning: daemon (pid %d) not killed\n", pid);
+	} else
 		printf("\tdaemon (pid %d) killed\n", pid);
+out:
+	seteuid(uid);
 }
 
 /*
  * Write a message into the status file.
  */
+static void
 upstat(msg)
 	char *msg;
 {
 	register int fd;
-	char statfile[BUFSIZ];
+	char statfile[MAXPATHLEN];
 
-	bp = pbuf;
-	if ((ST = pgetstr("st", &bp)) == NULL)
+	if (cgetstr(bp, "st", &ST) == -1)
 		ST = DEFSTAT;
-	(void) sprintf(statfile, "%s/%s", SD, ST);
+	(void) snprintf(statfile, sizeof(statfile), "%s/%s", SD, ST);
 	umask(0);
 	fd = open(statfile, O_WRONLY|O_CREAT, 0664);
 	if (fd < 0 || flock(fd, LOCK_EX) < 0) {
@@ -172,7 +228,9 @@ upstat(msg)
 /*
  * Remove all spool files and temporaries from the spooling area.
  */
+void
 clean(argc, argv)
+	int argc;
 	char *argv[];
 {
 	register int c, status;
@@ -185,10 +243,11 @@ clean(argc, argv)
 	}
 	if (argc == 2 && !strcmp(argv[1], "all")) {
 		printer = prbuf;
-		while (getprent(line) > 0) {
+		while (cgetnext(&bp, printcapdb) > 0) {
 			cp1 = prbuf;
-			cp2 = line;
-			while ((c = *cp2++) && c != '|' && c != ':')
+			cp2 = bp;
+			while ((c = *cp2++) && c != '|' && c != ':' &&
+			    (cp1 - prbuf) < sizeof(prbuf))
 				*cp1++ = c;
 			*cp1 = '\0';
 			cleanpr();
@@ -197,19 +256,22 @@ clean(argc, argv)
 	}
 	while (--argc) {
 		printer = *++argv;
-		if ((status = pgetent(line, printer)) < 0) {
+		if ((status = cgetent(&bp, printcapdb, printer)) == -2) {
 			printf("cannot open printer description file\n");
 			continue;
-		} else if (status == 0) {
+		} else if (status == -1) {
 			printf("unknown printer %s\n", printer);
 			continue;
-		}
+		} else if (status == -3)
+			fatal("potential reference loop detected in printcap file");
+
 		cleanpr();
 	}
 }
 
-xselect(d)
-struct direct *d;
+static int
+doselect(d)
+	struct dirent *d;
 {
 	int c = d->d_name[0];
 
@@ -222,12 +284,16 @@ struct direct *d;
  * Comparison routine for scandir. Sort by job number and machine, then
  * by `cf', `tf', or `df', then by the sequence letter A-Z, a-z.
  */
-sortq(d1, d2)
-struct direct **d1, **d2;
+static int
+sortq(a, b)
+	const void *a, *b;
 {
+	struct dirent **d1, **d2;
 	int c1, c2;
 
-	if (c1 = strcmp((*d1)->d_name + 3, (*d2)->d_name + 3))
+	d1 = (struct dirent **)a;
+	d2 = (struct dirent **)b;
+	if ((c1 = strcmp((*d1)->d_name + 3, (*d2)->d_name + 3)))
 		return(c1);
 	c1 = (*d1)->d_name[0];
 	c2 = (*d2)->d_name[0];
@@ -243,23 +309,25 @@ struct direct **d1, **d2;
 /*
  * Remove incomplete jobs from spooling area.
  */
+static void
 cleanpr()
 {
 	register int i, n;
 	register char *cp, *cp1, *lp;
-	struct direct **queue;
+	struct dirent **queue;
 	int nitems;
 
-	bp = pbuf;
-	if ((SD = pgetstr("sd", &bp)) == NULL)
+	if (cgetstr(bp, "sd", &SD) == -1)
 		SD = _PATH_DEFSPOOL;
 	printf("%s:\n", printer);
 
-	for (lp = line, cp = SD; *lp++ = *cp++; )
+	for (lp = line, cp = SD; (lp - line) < sizeof(line) && (*lp++ = *cp++);)
 		;
 	lp[-1] = '/';
 
-	nitems = scandir(SD, &queue, xselect, sortq);
+	seteuid(euid);
+	nitems = scandir(SD, &queue, doselect, sortq);
+	seteuid(uid);
 	if (nitems < 0) {
 		printf("\tcannot examine spool directory\n");
 		return;
@@ -279,7 +347,8 @@ cleanpr()
 				n++;
 			}
 			if (n == 0) {
-				strcpy(lp, cp);
+				strncpy(lp, cp, sizeof(line) - strlen(line) - 1);
+				line[sizeof(line) - 1] = '\0';
 				unlinkf(line);
 			}
 		} else {
@@ -288,25 +357,31 @@ cleanpr()
 			 * been skipped above) or a tf file (which can always
 			 * be removed).
 			 */
-			strcpy(lp, cp);
+			strncpy(lp, cp, sizeof(line) - strlen(line) - 1);
+			line[sizeof(line) - 1] = '\0';
 			unlinkf(line);
 		}
      	} while (++i < nitems);
 }
  
+static void
 unlinkf(name)
 	char	*name;
 {
+	seteuid(euid);
 	if (unlink(name) < 0)
 		printf("\tcannot remove %s\n", name);
 	else
 		printf("\tremoved %s\n", name);
+	seteuid(uid);
 }
 
 /*
  * Enable queuing to the printer (allow lpr's).
  */
+void
 enable(argc, argv)
+	int argc;
 	char *argv[];
 {
 	register int c, status;
@@ -319,10 +394,11 @@ enable(argc, argv)
 	}
 	if (argc == 2 && !strcmp(argv[1], "all")) {
 		printer = prbuf;
-		while (getprent(line) > 0) {
+		while (cgetnext(&bp, printcapdb) > 0) {
 			cp1 = prbuf;
-			cp2 = line;
-			while ((c = *cp2++) && c != '|' && c != ':')
+			cp2 = bp;
+			while ((c = *cp2++) && c != '|' && c != ':' &&
+			    (cp1 - prbuf) < sizeof(prbuf))
 				*cp1++ = c;
 			*cp1 = '\0';
 			enablepr();
@@ -331,44 +407,50 @@ enable(argc, argv)
 	}
 	while (--argc) {
 		printer = *++argv;
-		if ((status = pgetent(line, printer)) < 0) {
+		if ((status = cgetent(&bp, printcapdb, printer)) == -2) {
 			printf("cannot open printer description file\n");
 			continue;
-		} else if (status == 0) {
+		} else if (status == -1) {
 			printf("unknown printer %s\n", printer);
 			continue;
-		}
+		} else if (status == -3)
+			fatal("potential reference loop detected in printcap file");
+
 		enablepr();
 	}
 }
 
+static void
 enablepr()
 {
 	struct stat stbuf;
 
-	bp = pbuf;
-	if ((SD = pgetstr("sd", &bp)) == NULL)
+	if (cgetstr(bp, "sd", &SD) == -1)
 		SD = _PATH_DEFSPOOL;
-	if ((LO = pgetstr("lo", &bp)) == NULL)
+	if (cgetstr(bp, "lo", &LO) == -1)
 		LO = DEFLOCK;
-	(void) sprintf(line, "%s/%s", SD, LO);
+	(void) snprintf(line, sizeof(line), "%s/%s", SD, LO);
 	printf("%s:\n", printer);
 
 	/*
 	 * Turn off the group execute bit of the lock file to enable queuing.
 	 */
+	seteuid(euid);
 	if (stat(line, &stbuf) >= 0) {
 		if (chmod(line, stbuf.st_mode & 0767) < 0)
 			printf("\tcannot enable queuing\n");
 		else
 			printf("\tqueuing enabled\n");
 	}
+	seteuid(uid);
 }
 
 /*
  * Disable queuing.
  */
+void
 disable(argc, argv)
+	int argc;
 	char *argv[];
 {
 	register int c, status;
@@ -381,10 +463,11 @@ disable(argc, argv)
 	}
 	if (argc == 2 && !strcmp(argv[1], "all")) {
 		printer = prbuf;
-		while (getprent(line) > 0) {
+		while (cgetnext(&bp, printcapdb) > 0) {
 			cp1 = prbuf;
-			cp2 = line;
-			while ((c = *cp2++) && c != '|' && c != ':')
+			cp2 = bp;
+			while ((c = *cp2++) && c != '|' && c != ':' &&
+			    (cp1 - prbuf) < sizeof(prbuf))
 				*cp1++ = c;
 			*cp1 = '\0';
 			disablepr();
@@ -393,32 +476,35 @@ disable(argc, argv)
 	}
 	while (--argc) {
 		printer = *++argv;
-		if ((status = pgetent(line, printer)) < 0) {
+		if ((status = cgetent(&bp, printcapdb, printer)) == -2) {
 			printf("cannot open printer description file\n");
 			continue;
-		} else if (status == 0) {
+		} else if (status == -1) {
 			printf("unknown printer %s\n", printer);
 			continue;
-		}
+		} else if (status == -3)
+			fatal("potential reference loop detected in printcap file");
+
 		disablepr();
 	}
 }
 
+static void
 disablepr()
 {
 	register int fd;
 	struct stat stbuf;
 
-	bp = pbuf;
-	if ((SD = pgetstr("sd", &bp)) == NULL)
+	if (cgetstr(bp, "sd", &SD) == -1)
 		SD = _PATH_DEFSPOOL;
-	if ((LO = pgetstr("lo", &bp)) == NULL)
+	if (cgetstr(bp, "lo", &LO) == -1)
 		LO = DEFLOCK;
-	(void) sprintf(line, "%s/%s", SD, LO);
+	(void) snprintf(line, sizeof(line), "%s/%s", SD, LO);
 	printf("%s:\n", printer);
 	/*
 	 * Turn on the group execute bit of the lock file to disable queuing.
 	 */
+	seteuid(euid);
 	if (stat(line, &stbuf) >= 0) {
 		if (chmod(line, (stbuf.st_mode & 0777) | 010) < 0)
 			printf("\tcannot disable queuing\n");
@@ -431,16 +517,18 @@ disablepr()
 			(void) close(fd);
 			printf("\tqueuing disabled\n");
 		}
-		return;
 	} else
 		printf("\tcannot stat lock file\n");
+	seteuid(uid);
 }
 
 /*
  * Disable queuing and printing and put a message into the status file
  * (reason for being down).
  */
+void
 down(argc, argv)
+	int argc;
 	char *argv[];
 {
 	register int c, status;
@@ -453,10 +541,11 @@ down(argc, argv)
 	}
 	if (!strcmp(argv[1], "all")) {
 		printer = prbuf;
-		while (getprent(line) > 0) {
+		while (cgetnext(&bp, printcapdb) > 0) {
 			cp1 = prbuf;
-			cp2 = line;
-			while ((c = *cp2++) && c != '|' && c != ':')
+			cp2 = bp;
+			while ((c = *cp2++) && c != '|' && c != ':' &&
+			    (cp1 - prbuf) < sizeof(prbuf))
 				*cp1++ = c;
 			*cp1 = '\0';
 			putmsg(argc - 2, argv + 2);
@@ -464,17 +553,21 @@ down(argc, argv)
 		return;
 	}
 	printer = argv[1];
-	if ((status = pgetent(line, printer)) < 0) {
+	if ((status = cgetent(&bp, printcapdb, printer)) == -2) {
 		printf("cannot open printer description file\n");
 		return;
-	} else if (status == 0) {
+	} else if (status == -1) {
 		printf("unknown printer %s\n", printer);
 		return;
-	}
+	} else if (status == -3)
+			fatal("potential reference loop detected in printcap file");
+
 	putmsg(argc - 2, argv + 2);
 }
 
+static void
 putmsg(argc, argv)
+	int argc;
 	char **argv;
 {
 	register int fd;
@@ -482,19 +575,19 @@ putmsg(argc, argv)
 	char buf[1024];
 	struct stat stbuf;
 
-	bp = pbuf;
-	if ((SD = pgetstr("sd", &bp)) == NULL)
+	if (cgetstr(bp, "sd", &SD) == -1)
 		SD = _PATH_DEFSPOOL;
-	if ((LO = pgetstr("lo", &bp)) == NULL)
+	if (cgetstr(bp, "lo", &LO) == -1)
 		LO = DEFLOCK;
-	if ((ST = pgetstr("st", &bp)) == NULL)
+	if (cgetstr(bp, "st", &ST) == -1)
 		ST = DEFSTAT;
 	printf("%s:\n", printer);
 	/*
 	 * Turn on the group execute bit of the lock file to disable queuing and
 	 * turn on the owner execute bit of the lock file to disable printing.
 	 */
-	(void) sprintf(line, "%s/%s", SD, LO);
+	(void) snprintf(line, sizeof(line), "%s/%s", SD, LO);
+	seteuid(euid);
 	if (stat(line, &stbuf) >= 0) {
 		if (chmod(line, (stbuf.st_mode & 0777) | 0110) < 0)
 			printf("\tcannot disable queuing\n");
@@ -507,18 +600,21 @@ putmsg(argc, argv)
 			(void) close(fd);
 			printf("\tprinter and queuing disabled\n");
 		}
+		seteuid(uid);
 		return;
 	} else
 		printf("\tcannot stat lock file\n");
 	/*
 	 * Write the message into the status file.
 	 */
-	(void) sprintf(line, "%s/%s", SD, ST);
+	(void) snprintf(line, sizeof(line), "%s/%s", SD, ST);
 	fd = open(line, O_WRONLY|O_CREAT, 0664);
 	if (fd < 0 || flock(fd, LOCK_EX) < 0) {
 		printf("\tcannot create status file\n");
+		seteuid(uid);
 		return;
 	}
+	seteuid(uid);
 	(void) ftruncate(fd, 0);
 	if (argc <= 0) {
 		(void) write(fd, "\n", 1);
@@ -528,7 +624,7 @@ putmsg(argc, argv)
 	cp1 = buf;
 	while (--argc >= 0) {
 		cp2 = *argv++;
-		while (*cp1++ = *cp2++)
+		while ((cp1 - buf) < sizeof(buf) && (*cp1++ = *cp2++))
 			;
 		cp1[-1] = ' ';
 	}
@@ -541,7 +637,9 @@ putmsg(argc, argv)
 /*
  * Exit lpc
  */
+void
 quit(argc, argv)
+	int argc;
 	char *argv[];
 {
 	exit(0);
@@ -550,7 +648,9 @@ quit(argc, argv)
 /*
  * Kill and restart the daemon.
  */
+void
 restart(argc, argv)
+	int argc;
 	char *argv[];
 {
 	register int c, status;
@@ -563,10 +663,11 @@ restart(argc, argv)
 	}
 	if (argc == 2 && !strcmp(argv[1], "all")) {
 		printer = prbuf;
-		while (getprent(line) > 0) {
+		while (cgetnext(&bp, printcapdb) > 0) {
 			cp1 = prbuf;
-			cp2 = line;
-			while ((c = *cp2++) && c != '|' && c != ':')
+			cp2 = bp;
+			while ((c = *cp2++) && c != '|' && c != ':' &&
+			    (cp1 - prbuf) < sizeof(prbuf))
 				*cp1++ = c;
 			*cp1 = '\0';
 			abortpr(0);
@@ -576,13 +677,15 @@ restart(argc, argv)
 	}
 	while (--argc) {
 		printer = *++argv;
-		if ((status = pgetent(line, printer)) < 0) {
+		if ((status = cgetent(&bp, printcapdb, printer)) == -2) {
 			printf("cannot open printer description file\n");
 			continue;
-		} else if (status == 0) {
+		} else if (status == -1) {
 			printf("unknown printer %s\n", printer);
 			continue;
-		}
+		} else if (status == -3)
+			fatal("potential reference loop detected in printcap file");
+
 		abortpr(0);
 		startpr(0);
 	}
@@ -591,7 +694,9 @@ restart(argc, argv)
 /*
  * Enable printing on the specified printer and startup the daemon.
  */
-start(argc, argv)
+void
+startcmd(argc, argv)
+	int argc;
 	char *argv[];
 {
 	register int c, status;
@@ -604,10 +709,11 @@ start(argc, argv)
 	}
 	if (argc == 2 && !strcmp(argv[1], "all")) {
 		printer = prbuf;
-		while (getprent(line) > 0) {
+		while (cgetnext(&bp, printcapdb) > 0) {
 			cp1 = prbuf;
-			cp2 = line;
-			while ((c = *cp2++) && c != '|' && c != ':')
+			cp2 = bp;
+			while ((c = *cp2++) && c != '|' && c != ':' &&
+			    (cp1 - prbuf) < sizeof(prbuf))
 				*cp1++ = c;
 			*cp1 = '\0';
 			startpr(1);
@@ -616,32 +722,36 @@ start(argc, argv)
 	}
 	while (--argc) {
 		printer = *++argv;
-		if ((status = pgetent(line, printer)) < 0) {
+		if ((status = cgetent(&bp, printcapdb, printer)) == -2) {
 			printf("cannot open printer description file\n");
 			continue;
-		} else if (status == 0) {
+		} else if (status == -1) {
 			printf("unknown printer %s\n", printer);
 			continue;
-		}
+		} else if (status == -3)
+			fatal("potential reference loop detected in printcap file");
+
 		startpr(1);
 	}
 }
 
+static void
 startpr(enable)
+	int enable;
 {
 	struct stat stbuf;
 
-	bp = pbuf;
-	if ((SD = pgetstr("sd", &bp)) == NULL)
+	if (cgetstr(bp, "sd", &SD) == -1)
 		SD = _PATH_DEFSPOOL;
-	if ((LO = pgetstr("lo", &bp)) == NULL)
+	if (cgetstr(bp, "lo", &LO) == -1)
 		LO = DEFLOCK;
-	(void) sprintf(line, "%s/%s", SD, LO);
+	(void) snprintf(line, sizeof(line), "%s/%s", SD, LO);
 	printf("%s:\n", printer);
 
 	/*
 	 * Turn off the owner execute bit of the lock file to enable printing.
 	 */
+	seteuid(euid);
 	if (enable && stat(line, &stbuf) >= 0) {
 		if (chmod(line, stbuf.st_mode & (enable==2 ? 0666 : 0677)) < 0)
 			printf("\tcannot enable printing\n");
@@ -652,24 +762,28 @@ startpr(enable)
 		printf("\tcouldn't start daemon\n");
 	else
 		printf("\tdaemon started\n");
+	seteuid(uid);
 }
 
 /*
  * Print the status of each queue listed or all the queues.
  */
+void
 status(argc, argv)
+	int argc;
 	char *argv[];
 {
 	register int c, status;
 	register char *cp1, *cp2;
 	char prbuf[100];
 
-	if (argc == 1) {
+	if (argc == 1 || argc == 2 && !strcmp(argv[1], "all")) {
 		printer = prbuf;
-		while (getprent(line) > 0) {
+		while (cgetnext(&bp, printcapdb) > 0) {
 			cp1 = prbuf;
-			cp2 = line;
-			while ((c = *cp2++) && c != '|' && c != ':')
+			cp2 = bp;
+			while ((c = *cp2++) && c != '|' && c != ':' &&
+			    (cp1 - prbuf) < sizeof(prbuf))
 				*cp1++ = c;
 			*cp1 = '\0';
 			prstat();
@@ -678,13 +792,15 @@ status(argc, argv)
 	}
 	while (--argc) {
 		printer = *++argv;
-		if ((status = pgetent(line, printer)) < 0) {
+		if ((status = cgetent(&bp, printcapdb, printer)) == -2) {
 			printf("cannot open printer description file\n");
 			continue;
-		} else if (status == 0) {
+		} else if (status == -1) {
 			printf("unknown printer %s\n", printer);
 			continue;
-		}
+		} else if (status == -3)
+			fatal("potential reference loop detected in printcap file");
+
 		prstat();
 	}
 }
@@ -692,22 +808,22 @@ status(argc, argv)
 /*
  * Print the status of the printer queue.
  */
+static void
 prstat()
 {
 	struct stat stbuf;
 	register int fd, i;
-	register struct direct *dp;
+	register struct dirent *dp;
 	DIR *dirp;
 
-	bp = pbuf;
-	if ((SD = pgetstr("sd", &bp)) == NULL)
+	if (cgetstr(bp, "sd", &SD) == -1)
 		SD = _PATH_DEFSPOOL;
-	if ((LO = pgetstr("lo", &bp)) == NULL)
+	if (cgetstr(bp, "lo", &LO) == -1)
 		LO = DEFLOCK;
-	if ((ST = pgetstr("st", &bp)) == NULL)
+	if (cgetstr(bp, "st", &ST) == -1)
 		ST = DEFSTAT;
 	printf("%s:\n", printer);
-	(void) sprintf(line, "%s/%s", SD, LO);
+	(void) snprintf(line, sizeof(line), "%s/%s", SD, LO);
 	if (stat(line, &stbuf) >= 0) {
 		printf("\tqueuing is %s\n",
 			(stbuf.st_mode & 010) ? "disabled" : "enabled");
@@ -736,12 +852,12 @@ prstat()
 	fd = open(line, O_RDONLY);
 	if (fd < 0 || flock(fd, LOCK_SH|LOCK_NB) == 0) {
 		(void) close(fd);	/* unlocks as well */
-		printf("\tno daemon present\n");
+		printf("\tprinter idle\n");
 		return;
 	}
 	(void) close(fd);
 	putchar('\t');
-	(void) sprintf(line, "%s/%s", SD, ST);
+	(void) snprintf(line, sizeof(line), "%s/%s", SD, ST);
 	fd = open(line, O_RDONLY);
 	if (fd >= 0) {
 		(void) flock(fd, LOCK_SH);
@@ -755,7 +871,9 @@ prstat()
  * Stop the specified daemon after completing the current job and disable
  * printing.
  */
+void
 stop(argc, argv)
+	int argc;
 	char *argv[];
 {
 	register int c, status;
@@ -768,10 +886,11 @@ stop(argc, argv)
 	}
 	if (argc == 2 && !strcmp(argv[1], "all")) {
 		printer = prbuf;
-		while (getprent(line) > 0) {
+		while (cgetnext(&bp, printcapdb) > 0) {
 			cp1 = prbuf;
-			cp2 = line;
-			while ((c = *cp2++) && c != '|' && c != ':')
+			cp2 = bp;
+			while ((c = *cp2++) && c != '|' && c != ':' &&
+			    (cp1 - prbuf) < sizeof(prbuf))
 				*cp1++ = c;
 			*cp1 = '\0';
 			stoppr();
@@ -780,33 +899,36 @@ stop(argc, argv)
 	}
 	while (--argc) {
 		printer = *++argv;
-		if ((status = pgetent(line, printer)) < 0) {
+		if ((status = cgetent(&bp, printcapdb, printer)) == -2) {
 			printf("cannot open printer description file\n");
 			continue;
-		} else if (status == 0) {
+		} else if (status == -1) {
 			printf("unknown printer %s\n", printer);
 			continue;
-		}
+		} else if (status == -3)
+			fatal("potential reference loop detected in printcap file");
+
 		stoppr();
 	}
 }
 
+static void
 stoppr()
 {
 	register int fd;
 	struct stat stbuf;
 
-	bp = pbuf;
-	if ((SD = pgetstr("sd", &bp)) == NULL)
+	if (cgetstr(bp, "sd", &SD) == -1)
 		SD = _PATH_DEFSPOOL;
-	if ((LO = pgetstr("lo", &bp)) == NULL)
+	if (cgetstr(bp, "lo", &LO) == -1)
 		LO = DEFLOCK;
-	(void) sprintf(line, "%s/%s", SD, LO);
+	(void) snprintf(line, sizeof(line), "%s/%s", SD, LO);
 	printf("%s:\n", printer);
 
 	/*
 	 * Turn on the owner execute bit of the lock file to disable printing.
 	 */
+	seteuid(euid);
 	if (stat(line, &stbuf) >= 0) {
 		if (chmod(line, (stbuf.st_mode & 0777) | 0100) < 0)
 			printf("\tcannot disable printing\n");
@@ -824,6 +946,7 @@ stoppr()
 		}
 	} else
 		printf("\tcannot stat lock file\n");
+	seteuid(uid);
 }
 
 struct	queue **queue;
@@ -833,12 +956,13 @@ time_t	mtime;
 /*
  * Put the specified jobs at the top of printer queue.
  */
+void
 topq(argc, argv)
+	int argc;
 	char *argv[];
 {
-	register int n, i;
+	register int i;
 	struct stat stbuf;
-	register char *cfname;
 	int status, changed;
 
 	if (argc < 3) {
@@ -848,25 +972,28 @@ topq(argc, argv)
 
 	--argc;
 	printer = *++argv;
-	status = pgetent(line, printer);
-	if (status < 0) {
+	status = cgetent(&bp, printcapdb, printer);
+	if (status == -2) {
 		printf("cannot open printer description file\n");
 		return;
-	} else if (status == 0) {
+	} else if (status == -1) {
 		printf("%s: unknown printer\n", printer);
 		return;
-	}
-	bp = pbuf;
-	if ((SD = pgetstr("sd", &bp)) == NULL)
+	} else if (status == -3)
+		fatal("potential reference loop detected in printcap file");
+
+	if (cgetstr(bp, "sd", &SD) == -1)
 		SD = _PATH_DEFSPOOL;
-	if ((LO = pgetstr("lo", &bp)) == NULL)
+	if (cgetstr(bp, "lo", &LO) == -1)
 		LO = DEFLOCK;
 	printf("%s:\n", printer);
 
+	seteuid(euid);
 	if (chdir(SD) < 0) {
 		printf("\tcannot chdir to %s\n", SD);
-		return;
+		goto out;
 	}
+	seteuid(uid);
 	nitems = getq(&queue);
 	if (nitems == 0)
 		return;
@@ -890,28 +1017,38 @@ topq(argc, argv)
 	 * Turn on the public execute bit of the lock file to
 	 * get lpd to rebuild the queue after the current job.
 	 */
+	seteuid(euid);
 	if (changed && stat(LO, &stbuf) >= 0)
 		(void) chmod(LO, (stbuf.st_mode & 0777) | 01);
+
+out:
+	seteuid(uid);
 } 
 
 /*
  * Reposition the job by changing the modification time of
  * the control file.
  */
+static int
 touch(q)
 	struct queue *q;
 {
 	struct timeval tvp[2];
+	int ret;
 
 	tvp[0].tv_sec = tvp[1].tv_sec = --mtime;
 	tvp[0].tv_usec = tvp[1].tv_usec = 0;
-	return(utimes(q->q_name, tvp));
+	seteuid(euid);
+	ret = utimes(q->q_name, tvp);
+	seteuid(uid);
+	return (ret);
 }
 
 /*
  * Checks if specified job name is in the printer's queue.
  * Returns:  negative (-1) if argument name is not in the queue.
  */
+static int
 doarg(job)
 	char *job;
 {
@@ -925,7 +1062,7 @@ doarg(job)
 	 * Look for a job item consisting of system name, colon, number 
 	 * (example: ucbarpa:114)  
 	 */
-	if ((cp = index(job, ':')) != NULL) {
+	if ((cp = strchr(job, ':')) != NULL) {
 		machine = job;
 		*cp++ = '\0';
 		job = cp;
@@ -961,7 +1098,10 @@ doarg(job)
 	 * Process item consisting of owner's name (example: henry).
 	 */
 	for (qq = queue + nitems; --qq >= queue; ) {
-		if ((fp = fopen((*qq)->q_name, "r")) == NULL)
+		seteuid(euid);
+		fp = fopen((*qq)->q_name, "r");
+		seteuid(uid);
+		if (fp == NULL)
 			continue;
 		while (getline(fp) > 0)
 			if (line[0] == 'P')
@@ -980,7 +1120,9 @@ doarg(job)
 /*
  * Enable everything and start printer (undo `down').
  */
+void
 up(argc, argv)
+	int argc;
 	char *argv[];
 {
 	register int c, status;
@@ -993,10 +1135,11 @@ up(argc, argv)
 	}
 	if (argc == 2 && !strcmp(argv[1], "all")) {
 		printer = prbuf;
-		while (getprent(line) > 0) {
+		while (cgetnext(&bp, printcapdb) > 0) {
 			cp1 = prbuf;
-			cp2 = line;
-			while ((c = *cp2++) && c != '|' && c != ':')
+			cp2 = bp;
+			while ((c = *cp2++) && c != '|' && c != ':' &&
+			    (cp1 - prbuf) < sizeof(prbuf))
 				*cp1++ = c;
 			*cp1 = '\0';
 			startpr(2);
@@ -1005,13 +1148,15 @@ up(argc, argv)
 	}
 	while (--argc) {
 		printer = *++argv;
-		if ((status = pgetent(line, printer)) < 0) {
+		if ((status = cgetent(&bp, printcapdb, printer)) == -2) {
 			printf("cannot open printer description file\n");
 			continue;
-		} else if (status == 0) {
+		} else if (status == -1) {
 			printf("unknown printer %s\n", printer);
 			continue;
-		}
+		} else if (status == -3)
+			fatal("potential reference loop detected in printcap file");
+
 		startpr(2);
 	}
 }
