@@ -1,6 +1,9 @@
+/*	$OpenBSD: lpc.c,v 1.5 1997/01/17 16:12:37 millert Exp $	*/
+
 /*
- * Copyright (c) 1983 Regents of the University of California.
- * All rights reserved.
+ * Copyright (c) 1983, 1993
+ *	The Regents of the University of California.  All rights reserved.
+ *
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -32,40 +35,72 @@
  */
 
 #ifndef lint
-char copyright[] =
-"@(#) Copyright (c) 1983 Regents of the University of California.\n\
- All rights reserved.\n";
+static char copyright[] =
+"@(#) Copyright (c) 1983, 1993\n\
+	The Regents of the University of California.  All rights reserved.\n";
 #endif /* not lint */
 
+#ifndef lint
+#if 0
+static char sccsid[] = "@(#)lpc.c	8.3 (Berkeley) 4/28/95";
+#else
+static char rcsid[] = "$OpenBSD: lpc.c,v 1.5 1997/01/17 16:12:37 millert Exp $";
+#endif
+#endif /* not lint */
+
+#include <sys/param.h>
+
+#include <dirent.h>
+#include <signal.h>
+#include <setjmp.h>
+#include <syslog.h>
+#include <unistd.h>
+#include <stdlib.h>
+#include <stdio.h>
+#include <ctype.h>
+#include <string.h>
+#include <grp.h>
+#include <sys/param.h>
+#include "lp.h"
+#include "lpc.h"
+#include "extern.h"
+
+#ifndef LPR_OPER
+#define LPR_OPER	"operator"	/* group name of lpr operators */
+#endif
+
 /*
  * lpc -- line printer control program
  */
-#include <stdio.h>
-#include <signal.h>
-#include <ctype.h>
-#include <setjmp.h>
-#include <syslog.h>
-#include <string.h>
-
-#include "lpc.h"
 
+#define MAX_CMDLINE	200
+#define MAX_MARGV	20
 int	fromatty;
 
-char	cmdline[200];
+char	cmdline[MAX_CMDLINE];
 int	margc;
-char	*margv[20];
+char	*margv[MAX_MARGV];
 int	top;
-void	intr();
-struct	cmd *getcmd();
+uid_t	uid, euid;
 
 jmp_buf	toplevel;
 
+static void		 cmdscanner __P((int));
+static struct cmd	*getcmd __P((char *));
+static void		 intr __P((int));
+static void		 makeargv __P((void));
+static int		 ingroup __P((char *));
+
+int
 main(argc, argv)
+	int argc;
 	char *argv[];
 {
 	register struct cmd *c;
-	extern char *name;
 
+	euid = geteuid();
+	uid = getuid();
+	seteuid(uid);
 	name = argv[0];
 	openlog("lpd", 0, LOG_LPR);
 
@@ -79,7 +114,7 @@ main(argc, argv)
 			printf("?Invalid command\n");
 			exit(1);
 		}
-		if (c->c_priv && getuid()) {
+		if (c->c_priv && getuid() && ingroup(LPR_OPER) == 0) {
 			printf("?Privileged command\n");
 			exit(1);
 		}
@@ -96,8 +131,9 @@ main(argc, argv)
 	}
 }
 
-void
-intr()
+static void
+intr(signo)
+	int signo;
 {
 	if (!fromatty)
 		exit(0);
@@ -107,6 +143,7 @@ intr()
 /*
  * Command parser.
  */
+static void
 cmdscanner(top)
 	int top;
 {
@@ -119,8 +156,8 @@ cmdscanner(top)
 			printf("lpc> ");
 			fflush(stdout);
 		}
-		if (fgets(cmdline, sizeof(cmdline), stdin) == 0)
-			quit();
+		if (fgets(cmdline, MAX_CMDLINE, stdin) == 0)
+			quit(0, NULL);
 		if (cmdline[0] == 0 || cmdline[0] == '\n')
 			break;
 		makeargv();
@@ -133,7 +170,7 @@ cmdscanner(top)
 			printf("?Invalid command\n");
 			continue;
 		}
-		if (c->c_priv && getuid()) {
+		if (c->c_priv && getuid() && ingroup(LPR_OPER) == 0) {
 			printf("?Privileged command\n");
 			continue;
 		}
@@ -142,9 +179,7 @@ cmdscanner(top)
 	longjmp(toplevel, 0);
 }
 
-extern struct cmd cmdtab[];
-
-struct cmd *
+static struct cmd *
 getcmd(name)
 	register char *name;
 {
@@ -155,7 +190,7 @@ getcmd(name)
 	longest = 0;
 	nmatches = 0;
 	found = 0;
-	for (c = cmdtab; p = c->c_name; c++) {
+	for (c = cmdtab; (p = c->c_name); c++) {
 		for (q = name; *q == *p++; q++)
 			if (*q == 0)		/* exact match? */
 				return(c);
@@ -176,14 +211,16 @@ getcmd(name)
 /*
  * Slice a string up into argc/argv.
  */
+static void
 makeargv()
 {
 	register char *cp;
 	register char **argp = margv;
-	int n=0;
+	register int n = 0;
 
 	margc = 0;
-	for (cp = cmdline; *cp && (cp - cmdline) < sizeof(cmdline) && n < 20; n++) {
+	for (cp = cmdline; *cp && (cp - cmdline) < sizeof(cmdline) &&
+	    n < MAX_MARGV; n++) {
 		while (isspace(*cp))
 			cp++;
 		if (*cp == '\0')
@@ -204,6 +241,7 @@ makeargv()
 /*
  * Help command.
  */
+void
 help(argc, argv)
 	int argc;
 	char *argv[];
@@ -258,3 +296,33 @@ help(argc, argv)
 				c->c_name, c->c_help);
 	}
 }
+
+/*
+ * return non-zero if the user is a member of the given group
+ */
+static int
+ingroup(grname)
+	char *grname;
+{
+	static struct group *gptr=NULL;
+	static gid_t groups[NGROUPS];
+	register gid_t gid;
+	register int i;
+
+	if (gptr == NULL) {
+		if ((gptr = getgrnam(grname)) == NULL) {
+			fprintf(stderr, "Warning: unknown group '%s'\n",
+				grname);
+			return(0);
+		}
+		if (getgroups(NGROUPS, groups) < 0) {
+			perror("getgroups");
+			exit(1);
+		}
+	}
+	gid = gptr->gr_gid;
+	for (i = 0; i < NGROUPS; i++)
+		if (gid == groups[i])
+			return(1);
+	return(0);
+}
