@@ -1,6 +1,10 @@
+/*	$OpenBSD: lpd.c,v 1.16 1998/08/03 16:53:16 millert Exp $ */
+/*	$NetBSD: lpd.c,v 1.7 1996/04/24 14:54:06 mrg Exp $	*/
+
 /*
- * Copyright (c) 1983 Regents of the University of California.
- * All rights reserved.
+ * Copyright (c) 1983, 1993, 1994
+ *	The Regents of the University of California.  All rights reserved.
+ *
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -32,14 +36,17 @@
  */
 
 #ifndef lint
-char copyright[] =
-"@(#) Copyright (c) 1983 Regents of the University of California.\n\
- All rights reserved.\n";
+static char copyright[] =
+"@(#) Copyright (c) 1983, 1993, 1994\n\
+	The Regents of the University of California.  All rights reserved.\n";
 #endif /* not lint */
 
 #ifndef lint
-/*static char sccsid[] = "from: @(#)lpd.c	5.12 (Berkeley) 3/7/91";*/
-static char rcsid[] = "$Id: lpd.c,v 1.3 1993/11/10 04:36:34 cgd Exp $";
+#if 0
+static char sccsid[] = "@(#)lpd.c	8.7 (Berkeley) 5/10/95";
+#else
+static char rcsid[] = "$OpenBSD: lpd.c,v 1.16 1998/08/03 16:53:16 millert Exp $";
+#endif
 #endif /* not lint */
 
 /*
@@ -71,25 +78,67 @@ static char rcsid[] = "$Id: lpd.c,v 1.3 1993/11/10 04:36:34 cgd Exp $";
  *	   w/o help of lpq and lprm programs.
  */
 
+#include <sys/param.h>
+#include <sys/wait.h>
+#include <sys/types.h>
+#include <sys/socket.h>
+#include <sys/un.h>
+#include <sys/stat.h>
+#include <sys/file.h>
+#include <netinet/in.h>
+
+#include <netdb.h>
+#include <unistd.h>
+#include <syslog.h>
+#include <signal.h>
+#include <errno.h>
+#include <fcntl.h>
+#include <dirent.h>
+#include <stdio.h>
+#include <stdlib.h>
+#include <string.h>
+#include <ctype.h>
 #include "lp.h"
+#include "lp.local.h"
 #include "pathnames.h"
+#include "extern.h"
+
+extern int __ivaliduser __P((FILE *, in_addr_t, const char *, const char *));
 
 int	lflag;				/* log requests flag */
 int	from_remote;			/* from remote socket */
 
-void mcleanup(), reapchild();
+static void       reapchild __P((int));
+static void       mcleanup __P((int));
+static void       doit __P((void));
+static void       startup __P((void));
+static void       chkhost __P((struct sockaddr_in *));
+static int	  ckqueue __P((char *));
 
+uid_t	uid, euid;
+
+int
 main(argc, argv)
 	int argc;
 	char **argv;
 {
-	int f, funix, finet, options = 0, defreadfds, fromlen;
-	struct sockaddr_un sun, fromunix;
+	int f, funix, finet, options, fromlen;
+	fd_set defreadfds;
+	struct sockaddr_un un, fromunix;
 	struct sockaddr_in sin, frominet;
 	int omask, lfd;
 
+	euid = geteuid();	/* these shouldn't be different */
+	uid = getuid();
+	options = 0;
 	gethostname(host, sizeof(host));
-	name = argv[0];
+
+	name = "lpd";
+
+	if (euid != 0) {
+		fprintf(stderr,"lpd: must run as root\n");
+		exit(1);
+	}
 
 	while (--argc > 0) {
 		argv++;
@@ -112,6 +161,7 @@ main(argc, argv)
 #endif
 
 	openlog("lpd", LOG_PID, LOG_LPR);
+	syslog(LOG_INFO, "restarted");
 	(void) umask(0);
 	lfd = open(_PATH_MASTERLOCK, O_WRONLY|O_CREAT, 0644);
 	if (lfd < 0) {
@@ -134,7 +184,7 @@ main(argc, argv)
 		syslog(LOG_ERR, "%s: %m", _PATH_MASTERLOCK);
 		exit(1);
 	}
-	signal(SIGCHLD, SIG_IGN);
+	signal(SIGCHLD, reapchild);
 	/*
 	 * Restart all the printers.
 	 */
@@ -147,66 +197,82 @@ main(argc, argv)
 	}
 #define	mask(s)	(1 << ((s) - 1))
 	omask = sigblock(mask(SIGHUP)|mask(SIGINT)|mask(SIGQUIT)|mask(SIGTERM));
+	(void) umask(07);
 	signal(SIGHUP, mcleanup);
 	signal(SIGINT, mcleanup);
 	signal(SIGQUIT, mcleanup);
 	signal(SIGTERM, mcleanup);
-	bzero(&sun, sizeof(sun));
-	sun.sun_family = AF_UNIX;
-	strcpy(sun.sun_path, _PATH_SOCKETNAME);
-	if (bind(funix,
-	     (struct sockaddr *)&sun, strlen(sun.sun_path) + 2) < 0) {
+	memset(&un, 0, sizeof(un));
+	un.sun_family = AF_UNIX;
+	strcpy(un.sun_path, _PATH_SOCKETNAME);
+#ifndef SUN_LEN
+#define SUN_LEN(unp) (strlen((unp)->sun_path) + 2)
+#endif
+	if (bind(funix, (struct sockaddr *)&un, SUN_LEN(&un)) < 0) {
 		syslog(LOG_ERR, "ubind: %m");
 		exit(1);
 	}
+	(void) umask(0);
 	sigsetmask(omask);
-	defreadfds = 1 << funix;
+	FD_ZERO(&defreadfds);
+	FD_SET(funix, &defreadfds);
 	listen(funix, 5);
 	finet = socket(AF_INET, SOCK_STREAM, 0);
 	if (finet >= 0) {
 		struct servent *sp;
 
-		if (options & SO_DEBUG)
+		if (options & SO_DEBUG) {
 			if (setsockopt(finet, SOL_SOCKET, SO_DEBUG, 0, 0) < 0) {
 				syslog(LOG_ERR, "setsockopt (SO_DEBUG): %m");
-				mcleanup();
+				mcleanup(0);
 			}
+		}
+		f = 1;
+		(void) setsockopt(finet, SOL_SOCKET, SO_REUSEADDR, &f,
+		    sizeof(f));
 		sp = getservbyname("printer", "tcp");
 		if (sp == NULL) {
 			syslog(LOG_ERR, "printer/tcp: unknown service");
-			mcleanup();
+			mcleanup(0);
 		}
-		bzero(&sin, sizeof(sin));
+		memset(&sin, 0, sizeof(sin));
 		sin.sin_family = AF_INET;
-		sin.sin_addr.s_addr = INADDR_ANY;
 		sin.sin_port = sp->s_port;
 		if (bind(finet, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
 			syslog(LOG_ERR, "bind: %m");
-			mcleanup();
+			mcleanup(0);
 		}
-		defreadfds |= 1 << finet;
+		FD_SET(finet, &defreadfds);
 		listen(finet, 5);
 	}
 	/*
 	 * Main loop: accept, do a request, continue.
 	 */
+	memset(&frominet, 0, sizeof(frominet));
+	memset(&fromunix, 0, sizeof(fromunix));
 	for (;;) {
-		int domain, nfds, s, readfds = defreadfds;
+		int domain, nfds, s;
+		fd_set readfds;
 
+		FD_COPY(&defreadfds, &readfds);
 		nfds = select(20, &readfds, 0, 0, 0);
 		if (nfds <= 0) {
 			if (nfds < 0 && errno != EINTR)
 				syslog(LOG_WARNING, "select: %m");
 			continue;
 		}
-		if (readfds & (1 << funix)) {
+		if (FD_ISSET(funix, &readfds)) {
 			domain = AF_UNIX, fromlen = sizeof(fromunix);
 			s = accept(funix,
 			    (struct sockaddr *)&fromunix, &fromlen);
-		} else if (readfds & (1 << finet)) {
+		} else /* if (FD_ISSET(finet, &readfds)) */  {
 			domain = AF_INET, fromlen = sizeof(frominet);
 			s = accept(finet,
 			    (struct sockaddr *)&frominet, &fromlen);
+			if (frominet.sin_port == htons(20)) {
+				close(s);
+				continue;
+			}
 		}
 		if (s < 0) {
 			if (errno != EINTR)
@@ -235,17 +301,21 @@ main(argc, argv)
 	}
 }
 
-void
-reapchild()
+static void
+reapchild(signo)
+	int signo;
 {
-	union wait status;
+	int status;
+	int save_errno = errno;
 
-	while (wait3((int *)&status, WNOHANG, 0) > 0)
+	while (waitpid((pid_t)-1, &status, WNOHANG) > 0)
 		;
+	errno = save_errno;
 }
 
-void
-mcleanup()
+static void
+mcleanup(signo)
+	int signo;
 {
 	if (lflag)
 		syslog(LOG_INFO, "exiting");
@@ -262,8 +332,8 @@ int	requ[MAXREQUESTS];	/* job number of spool entries */
 int	requests;		/* # of spool requests */
 char	*person;		/* name of person doing lprm */
 
-char	fromb[32];	/* buffer for client's machine name */
-char	cbuf[BUFSIZ];	/* command line buffer */
+char	fromb[MAXHOSTNAMELEN];	/* buffer for client's machine name */
+char	cbuf[BUFSIZ];		/* command line buffer */
 char	*cmdnames[] = {
 	"null",
 	"printjob",
@@ -273,6 +343,7 @@ char	*cmdnames[] = {
 	"rmjob"
 };
 
+static void
 doit()
 {
 	register char *cp;
@@ -294,7 +365,7 @@ doit()
 		if (lflag) {
 			if (*cp >= '\1' && *cp <= '\5')
 				syslog(LOG_INFO, "%s requests %s %s",
-					from, cmdnames[*cp], cp+1);
+					from, cmdnames[(int)*cp], cp+1);
 			else
 				syslog(LOG_INFO, "bad request (%d) from %s",
 					*cp, from);
@@ -380,80 +451,111 @@ doit()
  * Make a pass through the printcap database and start printing any
  * files left from the last time the machine went down.
  */
+static void
 startup()
 {
-	char buf[BUFSIZ];
+	char *buf;
 	register char *cp;
 	int pid;
 
-	printer = buf;
-
 	/*
 	 * Restart the daemons.
 	 */
-	while (getprent(buf) > 0) {
+	while (cgetnext(&buf, printcapdb) > 0) {
+		if (ckqueue(buf) <= 0) {
+			free(buf);
+			continue;	/* no work to do for this printer */
+		}
 		for (cp = buf; *cp; cp++)
 			if (*cp == '|' || *cp == ':') {
 				*cp = '\0';
 				break;
 			}
+		if (lflag)
+			syslog(LOG_INFO, "work for %s", buf);
 		if ((pid = fork()) < 0) {
 			syslog(LOG_WARNING, "startup: cannot fork");
-			mcleanup();
+			mcleanup(0);
 		}
 		if (!pid) {
-			endprent();
+			printer = buf;
+			cgetclose();
 			printjob();
+			/* NOTREACHED */
 		}
+		else free(buf);
 	}
 }
 
+/*
+ * Make sure there's some work to do before forking off a child
+ */
+static int
+ckqueue(cap)
+	char *cap;
+{
+	register struct dirent *d;
+	DIR *dirp;
+	char *spooldir;
+
+	if (cgetstr(cap, "sd", &spooldir) == -1)
+		spooldir = _PATH_DEFSPOOL;
+	if ((dirp = opendir(spooldir)) == NULL)
+		return (-1);
+	while ((d = readdir(dirp)) != NULL) {
+		if (d->d_name[0] != 'c' || d->d_name[1] != 'f')
+			continue;	/* daemon control files only */
+		closedir(dirp);
+		return (1);		/* found something */
+	}
+	closedir(dirp);
+	return (0);
+}
+
 #define DUMMY ":nobody::"
 
 /*
  * Check to see if the from host has access to the line printer.
  */
+static void
 chkhost(f)
 	struct sockaddr_in *f;
 {
 	register struct hostent *hp;
 	register FILE *hostf;
-	register char *cp, *sp;
-	char ahost[50];
 	int first = 1;
-	extern char *inet_ntoa();
-	int baselen = -1;
+	int good = 0;
 
-	f->sin_port = ntohs(f->sin_port);
-	if (f->sin_family != AF_INET || f->sin_port >= IPPORT_RESERVED)
-		fatal("Malformed from address");
+	/* Need real hostname for temporary filenames */
 	hp = gethostbyaddr((char *)&f->sin_addr,
 	    sizeof(struct in_addr), f->sin_family);
-	if (hp == 0)
+	if (hp == NULL)
 		fatal("Host name for your address (%s) unknown",
 			inet_ntoa(f->sin_addr));
 
-	strcpy(fromb, hp->h_name);
+	(void) strncpy(fromb, hp->h_name, sizeof(fromb)-1);
+	from[sizeof(fromb) - 1] = '\0';
 	from = fromb;
-	if (!strcmp(from, host))
-		return;
 
-	sp = fromb;
-	cp = ahost;
-	while (*sp) {
-		if (*sp == '.') {
-			if (baselen == -1)
-				baselen = sp - fromb;
-			*cp++ = *sp++;
-		} else {
-			*cp++ = isupper(*sp) ? tolower(*sp++) : *sp++;
-		}
+	/* Check for spoof, ala rlogind */
+	hp = gethostbyname(fromb);
+	if (!hp)
+		fatal("hostname for your address (%s) unknown",
+		    inet_ntoa(f->sin_addr));
+	for (; good == 0 && hp->h_addr_list[0] != NULL; hp->h_addr_list++) {
+		if (!bcmp(hp->h_addr_list[0], (caddr_t)&f->sin_addr,
+		    sizeof(f->sin_addr)))
+			good = 1;
 	}
-	*cp = '\0';
+	if (good == 0)
+		fatal("address for your hostname (%s) not matched",
+		    inet_ntoa(f->sin_addr));
+
 	hostf = fopen(_PATH_HOSTSEQUIV, "r");
 again:
 	if (hostf) {
-		if (!_validuser(hostf, ahost, DUMMY, DUMMY, baselen)) {
+		if (__ivaliduser(hostf, f->sin_addr.s_addr,
+		    DUMMY, DUMMY) == 0) {
 			(void) fclose(hostf);
 			return;
 		}
@@ -465,4 +567,5 @@ again:
 		goto again;
 	}
 	fatal("Your host does not have line printer access");
+	/*NOTREACHED*/
 }
