@@ -1,6 +1,9 @@
+/*	$OpenBSD: recvjob.c,v 1.12 1997/10/05 11:37:01 deraadt Exp $	*/
+
 /*
- * Copyright (c) 1983 Regents of the University of California.
- * All rights reserved.
+ * Copyright (c) 1983, 1993
+ *	The Regents of the University of California.  All rights reserved.
+ *
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -32,49 +35,77 @@
  */
 
 #ifndef lint
-/*static char sccsid[] = "from: @(#)recvjob.c	5.15 (Berkeley) 5/4/91";*/
-static char rcsid[] = "$Id: recvjob.c,v 1.2 1993/08/01 17:58:49 mycroft Exp $";
+static char copyright[] =
+"@(#) Copyright (c) 1983, 1993\n\
+	The Regents of the University of California.  All rights reserved.\n";
+#endif /* not lint */
+
+#ifndef lint
+#if 0
+static char sccsid[] = "@(#)recvjob.c	8.2 (Berkeley) 4/27/95";
+#else
+static char rcsid[] = "$OpenBSD: recvjob.c,v 1.12 1997/10/05 11:37:01 deraadt Exp $";
+#endif
 #endif /* not lint */
 
 /*
  * Receive printer jobs from the network, queue them and
  * start the printer daemon.
  */
-
-#include "lp.h"
-#include "pathnames.h"
+#include <sys/param.h>
 #include <sys/mount.h>
-#ifdef __linux__
-#include <sys/vfs.h>
-#endif
+#include <sys/stat.h>
+
+#include <unistd.h>
+#include <signal.h>
+#include <fcntl.h>
+#include <dirent.h>
+#include <syslog.h>
+#include <stdio.h>
+#include <stdlib.h>
+#include <string.h>
+#include "lp.h"
+#include "lp.local.h"
+#include "extern.h"
+#include "pathnames.h"
 
-char	*sp = "";
 #define ack()	(void) write(1, sp, 1);
 
-char    tfname[40];		/* tmp copy of cf before linking */
-char    dfname[40];		/* data files */
-int	minfree;		/* keep at least minfree blocks available */
+static char	 dfname[NAME_MAX];	/* data files */
+static int	 minfree;       /* keep at least minfree blocks available */
+static char	*sp = "";
+static char	 tfname[NAME_MAX];	/* tmp copy of cf before linking */
 
-void	rcleanup();
+static int        chksize __P((int));
+static void       frecverr __P((const char *, ...));
+static int        noresponse __P((void));
+static void       rcleanup __P((int));
+static int        read_number __P((char *));
+static int        readfile __P((char *, int));
+static int        readjob __P((void));
 
+
+void
 recvjob()
 {
 	struct stat stb;
-	char *bp = pbuf;
 	int status;
 
 	/*
 	 * Perform lookup for printer name or abbreviation
 	 */
-	if ((status = pgetent(line, printer)) < 0)
+	if ((status = cgetent(&bp, printcapdb, printer)) == -2)
 		frecverr("cannot open printer description file");
-	else if (status == 0)
+	else if (status == -1)
 		frecverr("unknown printer %s", printer);
-	if ((LF = pgetstr("lf", &bp)) == NULL)
+	else if (status == -3)
+		fatal("potential reference loop detected in printcap file");
+	
+	if (cgetstr(bp, "lf", &LF) == -1)
 		LF = _PATH_CONSOLE;
-	if ((SD = pgetstr("sd", &bp)) == NULL)
+	if (cgetstr(bp, "sd", &SD) == -1)
 		SD = _PATH_DEFSPOOL;
-	if ((LO = pgetstr("lo", &bp)) == NULL)
+	if (cgetstr(bp, "lo", &LO) == -1)
 		LO = DEFLOCK;
 
 	(void) close(2);			/* set up log file */
@@ -105,6 +136,7 @@ recvjob()
  * Read printer jobs sent by lpd and copy them to the spooling directory.
  * Return the number of jobs successfully transfered.
  */
+static int
 readjob()
 {
 	register int size, nfiles;
@@ -120,15 +152,18 @@ readjob()
 		do {
 			if ((size = read(1, cp, 1)) != 1) {
 				if (size < 0)
-					frecverr("%s: Lost connection",printer);
+					frecverr("%s: Lost connection",
+					    printer);
 				return(nfiles);
 			}
-		} while (*cp++ != '\n');
+		} while (*cp++ != '\n' && (cp - line + 1) < sizeof line);
+		if (cp - line + 1 >= sizeof line)
+			frecverr("readjob overflow");
 		*--cp = '\0';
 		cp = line;
 		switch (*cp++) {
 		case '\1':	/* cleanup because data sent was bad */
-			rcleanup();
+			rcleanup(0);
 			continue;
 
 		case '\2':	/* read cf file */
@@ -143,15 +178,20 @@ readjob()
 			 * something different than what gethostbyaddr()
 			 * returns
 			 */
-			strcpy(cp + 6, from);
-			strcpy(tfname, cp);
+			strncpy(cp + 6, from, sizeof(line) + line - cp - 7);
+			line[sizeof(line) -1 ] = '\0';
+			strncpy(tfname, cp, sizeof tfname-1);
+			tfname[sizeof tfname-1] = '\0';
 			tfname[0] = 't';
+			if (strchr(tfname, '/'))
+				frecverr("readjob: %s: illegal path name",
+				    tfname);
 			if (!chksize(size)) {
 				(void) write(1, "\2", 1);
 				continue;
 			}
 			if (!readfile(tfname, size)) {
-				rcleanup();
+				rcleanup(0);
 				continue;
 			}
 			if (link(tfname, cp) < 0)
@@ -171,20 +211,22 @@ readjob()
 				(void) write(1, "\2", 1);
 				continue;
 			}
-			(void) strcpy(dfname, cp);
-			if (index(dfname, '/'))
+			(void) strncpy(dfname, cp, sizeof dfname-1);
+			dfname[sizeof dfname-1] = '\0';
+			if (strchr(dfname, '/'))
 				frecverr("readjob: %s: illegal path name",
 					dfname);
 			(void) readfile(dfname, size);
 			continue;
 		}
-		frecverr("protocol screwup");
+		frecverr("protocol screwup: %s", line);
 	}
 }
 
 /*
  * Read files send by lpd and copy them to the spooling directory.
  */
+static int
 readfile(file, size)
 	char *file;
 	int size;
@@ -223,13 +265,15 @@ readfile(file, size)
 	if (err)
 		frecverr("%s: write error", file);
 	if (noresponse()) {		/* file sent had bad data in it */
-		(void) unlink(file);
+		if (strchr(file, '/') == NULL)
+			(void) unlink(file);
 		return(0);
 	}
 	ack();
 	return(1);
 }
 
+static int
 noresponse()
 {
 	char resp;
@@ -245,6 +289,7 @@ noresponse()
  * Check to see if there is enough space on the disk for size bytes.
  * 1 == OK, 0 == Not OK.
  */
+static int
 chksize(size)
 	int size;
 {
@@ -262,6 +307,7 @@ chksize(size)
 	return(1);
 }
 
+static int
 read_number(fn)
 	char *fn;
 {
@@ -281,27 +327,49 @@ read_number(fn)
 /*
  * Remove all the files associated with the current job being transfered.
  */
-void
-rcleanup()
+static void
+rcleanup(signo)
+	int signo;
 {
-
-	if (tfname[0])
+	if (tfname[0] && strchr(tfname, '/') == NULL)
 		(void) unlink(tfname);
-	if (dfname[0])
+	if (dfname[0] && strchr(dfname, '/') == NULL) {
 		do {
 			do
 				(void) unlink(dfname);
 			while (dfname[2]-- != 'A');
 			dfname[2] = 'z';
 		} while (dfname[0]-- != 'd');
+	}
 	dfname[0] = '\0';
 }
 
-frecverr(msg, a1, a2)
+#ifdef __STDC__
+#include <stdarg.h>
+#else
+#include <varargs.h>
+#endif
+
+static void
+#ifdef __STDC__
+frecverr(const char *msg, ...)
+#else
+frecverr(msg, va_alist)
 	char *msg;
+        va_dcl
+#endif
 {
-	rcleanup();
-	syslog(LOG_ERR, msg, a1, a2);
+	extern char fromb[];
+	va_list ap;
+#ifdef __STDC__
+	va_start(ap, msg);
+#else
+	va_start(ap);
+#endif
+	rcleanup(0);
+	syslog(LOG_ERR, "%s", fromb);
+	vsyslog(LOG_ERR, msg, ap);
+	va_end(ap);
 	putchar('\1');		/* return error code */
 	exit(1);
 }
