@@ -1,6 +1,9 @@
+/*	$OpenBSD: lpq.c,v 1.7 1997/01/17 16:12:44 millert Exp $	*/
+
 /*
- * Copyright (c) 1983 Regents of the University of California.
- * All rights reserved.
+ * Copyright (c) 1983, 1993
+ *	The Regents of the University of California.  All rights reserved.
+ *
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -32,51 +35,77 @@
  */
 
 #ifndef lint
-char copyright[] =
-"@(#) Copyright (c) 1983 Regents of the University of California.\n\
- All rights reserved.\n";
+static char copyright[] =
+"@(#) Copyright (c) 1983, 1993\n\
+	The Regents of the University of California.  All rights reserved.\n";
 #endif /* not lint */
 
 #ifndef lint
-/*static char sccsid[] = "from: @(#)lpq.c	5.7 (Berkeley) 3/2/91";*/
-static char rcsid[] = "$Id: lpq.c,v 1.3 1999/10/17 18:38:17 notting Exp $";
+#if 0
+static char sccsid[] = "@(#)lpq.c	8.3 (Berkeley) 5/10/95";
+#else
+static char rcsid[] = "$OpenBSD: lpq.c,v 1.7 1997/01/17 16:12:44 millert Exp $";
+#endif
 #endif /* not lint */
 
 /*
  * Spool Queue examination program
  *
- * lpq [-l] [-Pprinter] [user...] [job...]
+ * lpq [-a] [-l] [-Pprinter] [user...] [job...]
  *
+ * -a show all non-null queues on the local machine
  * -l long output
  * -P used to identify printer as per lpr/lprm
  */
 
+#include <sys/param.h>
+
+#include <syslog.h>
+#include <dirent.h>
+#include <unistd.h>
+#include <stdlib.h>
+#include <stdio.h>
+#include <ctype.h>
 #include "lp.h"
+#include "lp.local.h"
+#include "pathnames.h"
 
+int	 requ[MAXREQUESTS];	/* job number of spool entries */
+int	 requests;		/* # of spool requests */
 char	*user[MAXUSERS];	/* users to process */
-int	users;			/* # of users in user array */
-int	requ[MAXREQUESTS];	/* job number of spool entries */
-int	requests;		/* # of spool requests */
+int	 users;			/* # of users in user array */
 
+uid_t	uid, euid;
+
+static int ckqueue __P((char *));
+void usage __P((void));
+
+int
 main(argc, argv)
 	register int	argc;
 	register char	**argv;
 {
 	extern char	*optarg;
 	extern int	optind;
-	int	ch, lflag;		/* long output option */
+	int	ch, aflag, lflag;
+	char	*buf, *cp;
 
+	euid = geteuid();
+	uid = getuid();
+	seteuid(uid);
 	name = *argv;
 	if (gethostname(host, sizeof(host))) {
 		perror("lpq: gethostname");
 		exit(1);
 	}
-	host[MAXHOSTNAMELEN-1]='\0';
 	openlog("lpd", 0, LOG_LPR);
 
-	lflag = 0;
-	while ((ch = getopt(argc, argv, "lP:")) != EOF)
+	aflag = lflag = 0;
+	while ((ch = getopt(argc, argv, "alP:")) != -1)
 		switch((char)ch) {
+		case 'a':
+			++aflag;
+			break;
 		case 'l':			/* long output */
 			++lflag;
 			break;
@@ -88,8 +117,13 @@ main(argc, argv)
 			usage();
 		}
 
-	if (printer == NULL && (printer = getenv("PRINTER")) == NULL)
+	if (!aflag && printer == NULL) {
+		char *p;
+
 		printer = DEFLP;
+		if ((p = getenv("PRINTER")) != NULL)
+			printer = p;
+	}
 
 	for (argc -= optind, argv += optind; argc; --argc, ++argv)
 		if (isdigit(argv[0][0])) {
@@ -103,12 +137,53 @@ main(argc, argv)
 			user[users++] = *argv;
 		}
 
-	displayq(lflag);
+	if (aflag) {
+		while (cgetnext(&buf, printcapdb) > 0) {
+			if (ckqueue(buf) <= 0) {
+				free(buf);
+				continue;	/* no jobs */
+			}
+			for (cp = buf; *cp; cp++)
+				if (*cp == '|' || *cp == ':') {
+					*cp = '\0';
+					break;
+				}
+			printer = buf;
+			printf("%s:\n", printer);
+			displayq(lflag);
+			free(buf);
+			printf("\n");
+		}
+	} else
+		displayq(lflag);
 	exit(0);
 }
 
+static int
+ckqueue(cap)
+	char *cap;
+{
+	register struct dirent *d;
+	DIR *dirp;
+	char *spooldir;
+
+	if (cgetstr(cap, "sd", &spooldir) == -1)
+		spooldir = _PATH_DEFSPOOL;
+	if ((dirp = opendir(spooldir)) == NULL)
+		return (-1);
+	while ((d = readdir(dirp)) != NULL) {
+		if (d->d_name[0] != 'c' || d->d_name[1] != 'f')
+			continue;	/* daemon control files only */
+		closedir(dirp);
+		return (1);		/* found something */
+	}
+	closedir(dirp);
+	return (0);
+}
+
+void
 usage()
 {
-	puts("usage: lpq [-l] [-Pprinter] [user ...] [job ...]");
+	puts("usage: lpq [-a] [-l] [-Pprinter] [user ...] [job ...]");
 	exit(1);
 }
