@@ -1,6 +1,15 @@
+/*	$OpenBSD: lpr.c,v 1.19 1997/07/25 20:12:13 mickey Exp $ */
+/*	$NetBSD: lpr.c,v 1.10 1996/03/21 18:12:25 jtc Exp $	*/
+
 /*
- * Copyright (c) 1983 Regents of the University of California.
- * All rights reserved.
+ * Copyright (c) 1983, 1989, 1993
+ *	The Regents of the University of California.  All rights reserved.
+ * (c) UNIX System Laboratories, Inc.
+ * All or some portions of this file are derived from material licensed
+ * to the University of California by American Telephone and Telegraph
+ * Co. or Unix System Laboratories, Inc. and are reproduced herein with
+ * the permission of UNIX System Laboratories, Inc.
+ *
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -32,15 +41,19 @@
  */
 
 #ifndef lint
-char copyright[] =
-"@(#) Copyright (c) 1983, 1989 Regents of the University of California.\n\
- All rights reserved.\n";
+static char copyright[] =
+"@(#) Copyright (c) 1983, 1989, 1993\n\
+	The Regents of the University of California.  All rights reserved.\n";
 #endif /* not lint */
 
 #ifndef lint
-/*static char sccsid[] = "from: @(#)lpr.c	5.9 (Berkeley) 3/2/91";*/
-static char rcsid[] = "$Id: lpr.c,v 1.4 1993/08/01 17:58:40 mycroft Exp $";
+#if 0
+static char sccsid[] = "@(#)lpr.c	8.4 (Berkeley) 4/28/95";
+#else
+static char rcsid[] = "$OpenBSD: lpr.c,v 1.19 1997/07/25 20:12:13 mickey Exp $";
+#endif
 #endif /* not lint */
+
 /*
  *      lpr -- off line print
  *
@@ -48,75 +61,80 @@ static char rcsid[] = "$Id: lpr.c,v 1.4 1993/08/01 17:58:40 mycroft Exp $";
  * using information from a printer data base.
  */
 
-#include <stdio.h>
-#include <sys/types.h>
-#include <sys/file.h>
+#include <sys/param.h>
 #include <sys/stat.h>
+#include <sys/file.h>
+
+#include <dirent.h>
+#include <fcntl.h>
+#include <a.out.h>
+#include <signal.h>
+#include <syslog.h>
 #include <pwd.h>
 #include <grp.h>
-#include <signal.h>
-#include <ctype.h>
-#include <syslog.h>
 #include <unistd.h>
 #include <stdlib.h>
+#include <stdio.h>
+#include <ctype.h>
 #include <string.h>
+#include "lp.h"
 #include "lp.local.h"
 #include "pathnames.h"
 
-char    *tfname;		/* tmp copy of cf before linking */
-char    *cfname;		/* daemon control files, linked from tf's */
-char    *dfname;		/* data files */
+static char	*cfname;	/* daemon control files, linked from tf's */
+static char	*class = host;	/* class title on header page */
+static char	*dfname;	/* data files */
+static char	*fonts[4];	/* troff font names */
+static char	 format = 'f';	/* format char for printing files */
+static int	 hdr = 1;	/* print header or not (default is yes) */
+static int	 iflag;		/* indentation wanted */
+static int	 inchar;	/* location to increment char in file names */
+static int	 indent;	/* amount to indent */
+static char	*jobname;	/* job name on header page */
+static int	 mailflg;	/* send mail */
+static int	 nact;		/* number of jobs to act on */
+static int	 ncopies = 1;	/* # of copies to make */
+static char	*person;	/* user name */
+static int	 qflag;		/* q job, but don't exec daemon */
+static int	 rflag;		/* remove files upon completion */	
+static int	 sflag;		/* symbolic link flag */
+static int	 tfd;		/* control file descriptor */
+static char	*tfname;	/* tmp copy of cf before linking */
+static char	*title;		/* pr'ing title */
+static int	 userid;	/* user id */
+static char	*width;		/* width for versatec printing */
 
-int	nact;			/* number of jobs to act on */
-int	tfd;			/* control file descriptor */
-int     mailflg;		/* send mail */
-int	qflag;			/* q job, but don't exec daemon */
-char	format = 'f';		/* format char for printing files */
-int	rflag;			/* remove files upon completion */	
-int	sflag;			/* symbolic link flag */
-int	inchar;			/* location to increment char in file names */
-int     ncopies = 1;		/* # of copies to make */
-int	iflag;			/* indentation wanted */
-int	indent;			/* amount to indent */
-int	hdr = 1;		/* print header or not (default is yes) */
-int     userid;			/* user id */
-char	*person;		/* user name */
-char	*title;			/* pr'ing title */
-char	*fonts[4];		/* troff font names */
-char	*width;			/* width for versatec printing */
-char	host[32];		/* host name */
-char	*class = host;		/* class title on header page */
-char    *jobname;		/* job name on header page */
-char	*name;			/* program name */
-char	*printer;		/* printer name */
-struct	stat statb;
+static struct stat statb;
 
-int	MX;			/* maximum number of blocks to copy */
-int	MC;			/* maximum number of copies allowed */
-int	DU;			/* daemon user-id */
-char	*SD;			/* spool directory */
-char	*LO;			/* lock file name */
-char	*RG;			/* restrict group */
-static short	SC;			/* suppress multiple copies */
+static void	 card __P((int, char *));
+static void	 chkprinter __P((char *));
+static void	 cleanup __P((int));
+static void	 copy __P((int, char []));
+static void	 fatal2 __P((const char *, ...));
+static char	*linked __P((char *));
+static char	*lmktemp __P((char *, int, int));
+static void	 mktemps __P((void));
+static int	 nfile __P((char *));
+static int	 test __P((char *));
+static char	*itoa __P((int));
 
-char	*getenv();
-char	*rindex();
-char	*linked();
-void	cleanup();
+uid_t	uid, euid;
 
-/*ARGSUSED*/
+int
 main(argc, argv)
 	int argc;
 	char *argv[];
 {
 	struct passwd *pw;
 	struct group *gptr;
-	extern char *itoa();
 	register char *arg, *cp;
 	char buf[BUFSIZ];
 	int i, f;
 	struct stat stb;
 
+	euid = geteuid();
+	uid = getuid();
+	seteuid(uid);
 	if (signal(SIGHUP, SIG_IGN) != SIG_IGN)
 		signal(SIGHUP, cleanup);
 	if (signal(SIGINT, SIG_IGN) != SIG_IGN)
@@ -245,13 +263,19 @@ main(argc, argv)
 			}
 		}
 	}
-	if (printer == NULL && (printer = getenv("PRINTER")) == NULL)
+
+	if (printer == NULL) {
+		char *p;
+
 		printer = DEFLP;
+		if ((p = getenv("PRINTER")) != NULL)
+			printer = p;
+	}
 	chkprinter(printer);
 	if (SC && ncopies > 1)
-		fatal("multiple copies are not allowed");
+		fatal2("multiple copies are not allowed");
 	if (MC > 0 && ncopies > MC)
-		fatal("only %d copies are allowed", MC);
+		fatal2("only %d copies are allowed", MC);
 	/*
 	 * Get the identity of the person doing the lpr using the same
 	 * algorithm as lprm. 
@@ -259,7 +283,7 @@ main(argc, argv)
 	userid = getuid();
 	if (userid != DU || person == 0) {
 		if ((pw = getpwuid(userid)) == NULL)
-			fatal("Who are you?");
+			fatal2("Who are you?");
 		person = pw->pw_name;
 	}
 	/*
@@ -267,7 +291,7 @@ main(argc, argv)
 	 */
 	if (RG != NULL && userid != DU) {
 		if ((gptr = getgrnam(RG)) == NULL)
-			fatal("Restricted group specified incorrectly");
+			fatal2("Restricted group specified incorrectly");
 		if (gptr->gr_gid != getgid()) {
 			while (*gptr->gr_mem != NULL) {
 				if ((strcmp(person, *gptr->gr_mem)) == 0)
@@ -275,21 +299,23 @@ main(argc, argv)
 				gptr->gr_mem++;
 			}
 			if (*gptr->gr_mem == NULL)
-				fatal("Not a member of the restricted group");
+				fatal2("Not a member of the restricted group");
 		}
 	}
 	/*
 	 * Check to make sure queuing is enabled if userid is not root.
 	 */
-	(void) sprintf(buf, "%s/%s", SD, LO);
+	(void) snprintf(buf, sizeof(buf), "%s/%s", SD, LO);
 	if (userid && stat(buf, &stb) == 0 && (stb.st_mode & 010))
-		fatal("Printer queue is disabled");
+		fatal2("Printer queue is disabled");
 	/*
 	 * Initialize the control file.
 	 */
 	mktemps();
 	tfd = nfile(tfname);
+	seteuid(euid);
 	(void) fchown(tfd, DU, -1);	/* owned by daemon for protection */
+	seteuid(uid);
 	card('H', host);
 	card('P', person);
 	if (hdr) {
@@ -297,7 +323,7 @@ main(argc, argv)
 			if (argc == 1)
 				jobname = "stdin";
 			else
-				jobname = (arg = rindex(argv[1], '/')) ? arg+1 : argv[1];
+				jobname = (arg = strrchr(argv[1], '/')) ? arg+1 : argv[1];
 		}
 		card('J', jobname);
 		card('C', class);
@@ -324,7 +350,8 @@ main(argc, argv)
 			continue;	/* file unreasonable */
 
 		if (sflag && (cp = linked(arg)) != NULL) {
-			(void) sprintf(buf, "%d %d", statb.st_dev, statb.st_ino);
+			(void) snprintf(buf, sizeof(buf), "%d %d", statb.st_dev,
+					statb.st_ino);
 			card('S', buf);
 			if (format == 'p')
 				card('T', title ? title : arg);
@@ -342,12 +369,12 @@ main(argc, argv)
 			printf("%s: %s: not linked, copying instead\n", name, arg);
 		if ((i = open(arg, O_RDONLY)) < 0) {
 			printf("%s: cannot open %s\n", name, arg);
-			continue;
+		} else {
+			copy(i, arg);
+			(void) close(i);
+			if (f && unlink(arg) < 0)
+				printf("%s: %s: not removed\n", name, arg);
 		}
-		copy(i, arg);
-		(void) close(i);
-		if (f && unlink(arg) < 0)
-			printf("%s: %s: not removed\n", name, arg);
 	}
 
 	if (nact) {
@@ -356,36 +383,41 @@ main(argc, argv)
 		/*
 		 * Touch the control file to fix position in the queue.
 		 */
+		seteuid(euid);
 		if ((tfd = open(tfname, O_RDWR)) >= 0) {
 			char c;
 
-			if (read(tfd, &c, 1) == 1 && lseek(tfd, 0L, 0) == 0 &&
+			if (read(tfd, &c, 1) == 1 &&
+			    lseek(tfd, (off_t)0, 0) == 0 &&
 			    write(tfd, &c, 1) != 1) {
 				printf("%s: cannot touch %s\n", name, tfname);
 				tfname[inchar]++;
-				cleanup();
+				cleanup(0);
 			}
 			(void) close(tfd);
 		}
 		if (link(tfname, cfname) < 0) {
 			printf("%s: cannot rename %s\n", name, cfname);
 			tfname[inchar]++;
-			cleanup();
+			cleanup(0);
 		}
 		unlink(tfname);
+		seteuid(uid);
 		if (qflag)		/* just q things up */
 			exit(0);
 		if (!startdaemon(printer))
 			printf("jobs queued, but cannot start daemon.\n");
 		exit(0);
 	}
-	cleanup();
+	cleanup(0);
+	return (1);
 	/* NOTREACHED */
 }
 
 /*
  * Create the file n and copy from file descriptor f.
  */
+static void
 copy(f, n)
 	int f;
 	char n[];
@@ -411,7 +443,8 @@ copy(f, n)
 			nc -= BUFSIZ;
 			nr++;
 			if (MX > 0 && nr > MX) {
-				printf("%s: %s: copy file is too large\n", name, n);
+				printf("%s: %s: copy file is too large\n",
+				       name, n);
 				break;
 			}
 		}
@@ -427,16 +460,18 @@ copy(f, n)
  * Try and link the file to dfname. Return a pointer to the full
  * path name if successful.
  */
-char *
+static char *
 linked(file)
 	register char *file;
 {
 	register char *cp;
-	static char buf[BUFSIZ];
+	static char nfile[MAXPATHLEN];
+	register int ret;
 
 	if (*file != '/') {
-		if (getwd(buf) == NULL)
+		if (getcwd(nfile, sizeof(nfile)) == NULL)
 			return(NULL);
+
 		while (file[0] == '.') {
 			switch (file[1]) {
 			case '/':
@@ -444,7 +479,7 @@ linked(file)
 				continue;
 			case '.':
 				if (file[2] == '/') {
-					if ((cp = rindex(buf, '/')) != NULL)
+					if ((cp = strrchr(nfile, '/')) != NULL)
 						*cp = '\0';
 					file += 3;
 					continue;
@@ -452,25 +487,30 @@ linked(file)
 			}
 			break;
 		}
-		strcat(buf, "/");
-		strcat(buf, file);
-		file = buf;
+		strncat(nfile, "/", sizeof(nfile) - strlen(nfile) - 1);
+		strncat(nfile, file, sizeof(nfile) - strlen(nfile) - 1);
+		file = nfile;
 	}
-	return(symlink(file, dfname) ? NULL : file);
+	seteuid(euid);
+	ret = symlink(file, dfname);
+	seteuid(uid);
+	return(ret ? NULL : file);
 }
 
 /*
  * Put a line into the control file.
  */
+static void
 card(c, p2)
-	register char c, *p2;
+	register int c;
+	register char *p2;
 {
 	char buf[BUFSIZ];
 	register char *p1 = buf;
 	register int len = 2;
 
 	*p1++ = c;
-	while ((c = *p2++) != '\0') {
+	while ((c = *p2++) != '\0' && len < sizeof(buf)) {
 		*p1++ = (c == '\n') ? ' ' : c;
 		len++;
 	}
@@ -481,26 +521,29 @@ card(c, p2)
 /*
  * Create a new file in the spool directory.
  */
+static int
 nfile(n)
 	char *n;
 {
-	register f;
+	register int f;
 	int oldumask = umask(0);		/* should block signals */
 
-	f = creat(n, FILMOD);
+	seteuid(euid);
+	f = open(n, O_WRONLY | O_EXCL | O_CREAT, FILMOD);
 	(void) umask(oldumask);
 	if (f < 0) {
 		printf("%s: cannot create %s\n", name, n);
-		cleanup();
+		cleanup(0);
 	}
 	if (fchown(f, userid, -1) < 0) {
 		printf("%s: cannot chown %s\n", name, n);
-		cleanup();
+		cleanup(0);	/* cleanup does exit */
 	}
+	seteuid(uid);
 	if (++n[inchar] > 'z') {
 		if (++n[inchar-2] == 't') {
 			printf("too many files - break up the job\n");
-			cleanup();
+			cleanup(0);
 		}
 		n[inchar] = 'A';
 	} else if (n[inchar] == '[')
@@ -511,8 +554,9 @@ nfile(n)
 /*
  * Cleanup after interrupts and errors.
  */
-void
-cleanup()
+static void
+cleanup(signo)
+	int signo;
 {
 	register i;
 
@@ -521,6 +565,7 @@ cleanup()
 	signal(SIGQUIT, SIG_IGN);
 	signal(SIGTERM, SIG_IGN);
 	i = inchar;
+	seteuid(euid);
 	if (tfname)
 		do
 			unlink(tfname);
@@ -544,6 +589,7 @@ cleanup()
  * Return -1 if it is not, 0 if its printable, and 1 if
  * we should remove it after printing.
  */
+static int
 test(file)
 	char *file;
 {
@@ -551,82 +597,56 @@ test(file)
 	register int fd;
 	register char *cp;
 
-	if (access(file, 4) < 0) {
-		printf("%s: cannot access %s\n", name, file);
-		return(-1);
+	if ((fd = open(file, O_RDONLY)) < 0) {
+		printf("%s: cannot open %s\n", name, file);
+		goto bad;
 	}
-	if (stat(file, &statb) < 0) {
+	if (fstat(fd, &statb) < 0) {
 		printf("%s: cannot stat %s\n", name, file);
-		return(-1);
+		goto bad;
 	}
 	if ((statb.st_mode & S_IFMT) == S_IFDIR) {
 		printf("%s: %s is a directory\n", name, file);
-		return(-1);
+		goto bad;
 	}
 	if (statb.st_size == 0) {
 		printf("%s: %s is an empty file\n", name, file);
-		return(-1);
+		goto bad;
  	}
-	if ((fd = open(file, O_RDONLY)) < 0) {
-		printf("%s: cannot open %s\n", name, file);
-		return(-1);
+	if (read(fd, &execb, sizeof(execb)) == sizeof(execb) &&
+	    !N_BADMAG(execb)) {
+			printf("%s: %s is an executable program and is unprintable",
+				name, file);
+			(void) close(fd);
+			goto bad;
 	}
-#ifndef	__linux__
-	if (read(fd, &execb, sizeof(execb)) == sizeof(execb)) {
-		if(!N_BADMAG(execb) || 
-#ifdef SARMAG
-		    strncmp((char *)&execb, ARMAG, SARMAG)==0
-#else
-		    execb.a_magic==ARMAG
-#endif
-		    ) {
-			printf("%s: %s is an executable program", name, file);
-			goto error1;
-		}
-#ifdef A_MAGIC1
-		switch(execb.a_magic) {
-		case A_MAGIC1:
-		case A_MAGIC2:
-		case A_MAGIC3:
-#ifdef A_MAGIC4
-		case A_MAGIC4:
-#endif
-			printf("%s: %s is an executable program", name, file);
-			goto error1;
-
-		case ARMAG:
-			printf("%s: %s is an archive file", name, file);
-			goto error1;
-		}
-#endif
-	}
-#endif
 	(void) close(fd);
 	if (rflag) {
-		if ((cp = rindex(file, '/')) == NULL) {
+		if ((cp = strrchr(file, '/')) == NULL) {
 			if (access(".", 2) == 0)
 				return(1);
 		} else {
-			*cp = '\0';
-			fd = access(file, 2);
-			*cp = '/';
+			if (cp == file) {
+				fd = access("/", 2);
+			} else {
+				*cp = '\0';
+				fd = access(file, 2);
+				*cp = '/';
+			}
 			if (fd == 0)
 				return(1);
 		}
 		printf("%s: %s: is not removable by you\n", name, file);
 	}
 	return(0);
-
-error1:
-	printf(" and is unprintable\n");
-	(void) close(fd);
+bad:
 	return(-1);
 }
 
 /*
  * itoa - integer to string conversion
  */
-char *
+static char *
 itoa(i)
 	register int i;
 {
@@ -643,43 +663,43 @@ itoa(i)
 /*
  * Perform lookup for printer name or abbreviation --
  */
+static void
 chkprinter(s)
 	char *s;
 {
 	int status;
-	char buf[BUFSIZ];
-	static char pbuf[BUFSIZ/2];
-	char *bp = pbuf;
-	extern char *pgetstr();
 
-	if ((status = pgetent(buf, s)) < 0)
-		fatal("cannot open printer description file");
-	else if (status == 0)
-		fatal("%s: unknown printer", s);
-	if ((SD = pgetstr("sd", &bp)) == NULL)
+	if ((status = cgetent(&bp, printcapdb, s)) == -2)
+		fatal2("cannot open printer description file");
+	else if (status == -1)
+		fatal2("%s: unknown printer", s);
+	if (cgetstr(bp, "sd", &SD) == -1)
 		SD = _PATH_DEFSPOOL;
-	if ((LO = pgetstr("lo", &bp)) == NULL)
+	if (cgetstr(bp, "lo", &LO) == -1)
 		LO = DEFLOCK;
-	RG = pgetstr("rg", &bp);
-	if ((MX = pgetnum("mx")) < 0)
+	cgetstr(bp, "rg", &RG);
+	if (cgetnum(bp, "mx", &MX) < 0)
 		MX = DEFMX;
-	if ((MC = pgetnum("mc")) < 0)
+	if (cgetnum(bp,"mc", &MC) < 0)
 		MC = DEFMAXCOPIES;
-	if ((DU = pgetnum("du")) < 0)
+	if (cgetnum(bp, "du", &DU) < 0)
 		DU = DEFUID;
-	SC = pgetflag("sc");
+	SC = (cgetcap(bp, "sc", ':') != NULL);
 }
 
 /*
  * Make the temp files.
  */
+static void
 mktemps()
 {
-	register int c, len, fd, n;
+	register int len, fd, n;
 	register char *cp;
 	char buf[BUFSIZ];
+	char *lmktemp();
 
-	(void) sprintf(buf, "%s/.seq", SD);
+	(void) snprintf(buf, sizeof(buf), "%s/.seq", SD);
+	seteuid(euid);
 	if ((fd = open(buf, O_RDWR|O_CREAT, 0661)) < 0) {
 		printf("%s: cannot create %s\n", name, buf);
 		exit(1);
@@ -688,6 +708,7 @@ mktemps()
 		printf("%s: cannot lock %s\n", name, buf);
 		exit(1);
 	}
+	seteuid(uid);
 	n = 0;
 	if ((len = read(fd, buf, sizeof(buf))) > 0) {
 		for (cp = buf; len--; ) {
@@ -696,30 +717,58 @@ mktemps()
 			n = n * 10 + (*cp++ - '0');
 		}
 	}
-	len = strlen(SD) + 15;
-	if ((cfname = malloc(len)) == NULL)
-		fatal("out of memory");
-	(void) sprintf(cfname, "%s/cfA%03dAXXXXXX", SD, n);
-	if (mktemp(cfname) == NULL)
-		fatal("couldn't create unique control file");
-	if ((tfname = strdup(cfname)) == NULL || (dfname = strdup(cfname)) == NULL)
-		fatal("out of memory");
+	len = strlen(SD) + strlen(host) + 8;
+	tfname = lmktemp("tf", n, len);
+	cfname = lmktemp("cf", n, len);
+	dfname = lmktemp("df", n, len);
 	inchar = strlen(SD) + 3;
-	tfname[inchar-2] = 't';
-	dfname[inchar-2] = 'd';
 	n = (n + 1) % 1000;
-	(void) lseek(fd, 0L, 0);
-	sprintf(buf, "%03d\n", n);
+	(void) lseek(fd, (off_t)0, 0);
+	snprintf(buf, sizeof(buf), "%03d\n", n);
 	(void) write(fd, buf, strlen(buf));
 	(void) close(fd);	/* unlocks as well */
 }
 
-/*VARARGS1*/
-fatal(msg, a1, a2, a3)
-	char *msg;
+/*
+ * Make a temp file name.
+ */
+static char *
+lmktemp(id, num, len)
+	char	*id;
+	int	num, len;
 {
+	register char *s;
+
+	if ((s = malloc(len)) == NULL)
+		fatal2("out of memory");
+	(void) snprintf(s, len, "%s/%sA%03d%s", SD, id, num, host);
+	return(s);
+}
+
+#ifdef __STDC__
+#include <stdarg.h>
+#else
+#include <varargs.h>
+#endif
+
+static void
+#ifdef __STDC__
+fatal2(const char *msg, ...)
+#else
+fatal2(msg, va_alist)
+	char *msg;
+        va_dcl
+#endif
+{
+	va_list ap;
+#ifdef __STDC__
+	va_start(ap, msg);
+#else
+	va_start(ap);
+#endif
 	printf("%s: ", name);
-	printf(msg, a1, a2, a3);
+	vprintf(msg, ap);
 	putchar('\n');
+	va_end(ap);
 	exit(1);
 }
